/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
 */
#include <media/custmcam.h>

#include <linux/delay.h>
#include <linux/types.h>
#include <linux/i2c.h>
#include <linux/uaccess.h>
#include <linux/miscdevice.h>
#include <media/msm_camera.h>
#include <mach/gpio.h>
#include <linux/slab.h>
#include "camsensor_drv.h"

#include <linux/sched.h>
#include <linux/timer.h>

#include <linux/time.h>

#include <mach/vreg.h>
#include <mach/pmic.h>

#include <linux/gpio.h>
#include <mach/gpio.h>

#include <mach/hs_io_ctl_a.h>
#include <linux/mfd/tps65023.h>

#include <mach/msm_battery.h>

#define SENSOR_ERROR_CHK_LOG
#undef SENSOR_ERROR_CHK_LOG

#define SENSOR_DEBUG_LOG
#undef SENSOR_DEBUG_LOG

#define SENSOR_REG_LOG
#undef SENSOR_REG_LOG

#ifdef SENSOR_SEQ_LOG
#include "linux/oem_msg.h"
#define CAM_SEQ_LOG(a,b,c,d) MSG_FATAL(a,b,c,d)
#else
#define CAM_SEQ_LOG(a,b,c,d)
#endif

#ifdef SENSOR_DEBUG_LOG
#include "linux/oem_msg.h"
#undef CDBG
#define CDBG(a,b,c,d) MSG_FATAL(a,b,c,d)
#else
#undef CDBG
#define CDBG(a,b,c,d)
#endif

#ifdef SENSOR_REG_LOG
#include "linux/oem_msg.h"
#define REG_DBG(a,b,c,d) MSG_FATAL(a,b,c,d)
#else
#define REG_DBG(a,b,c,d)
#endif

#ifdef SENSOR_ERROR_CHK_LOG
#include "linux/oem_msg.h"
#endif

#include "linux/oem_msg.h"
#define API_LOG_SWITCH
#undef API_LOG_SWITCH
#ifdef API_LOG_SWITCH
#define API_LOG(fmt, args...) printk(KERN_INFO fmt, ##args)
#else
#define API_LOG(fmt, args...) do{ } while(0)
#endif

#define DBG_LOG_SWITCH
#undef DBG_LOG_SWITCH
#ifdef DBG_LOG_SWITCH
#define DBG_LOG(fmt, args...) printk(KERN_NOTICE fmt, ##args)
#else
#define DBG_LOG(fmt, args...) do{ } while(0)
#endif

#define ERR_LOG_SWITCH

#ifdef ERR_LOG_SWITCH
#define ERR_LOG(fmt, args...) printk(KERN_ERR fmt, ##args)
#else
#define ERR_LOG(fmt, args...) do{ } while(0)
#endif

#define TIME_CHECK
#undef TIME_CHECK

#define _FLASH_ENABLE_

#define FLASH_TIMING_DEBUG
#undef FLASH_TIMING_DEBUG

#define I2C_DEBUG
#undef I2C_DEBUG

#define SENSOR_STATE_MODE        0x000000FF
#define SENSOR_STATE_FLASH       0x00000F00
#define SENSOR_STATE_AF          0x000FF000

#define SENSOR_OFF               0x00000000
#define SENSOR_ON                0x00000001
#define SENSOR_PREVIEW           0x00000002
#define SENSOR_STILL             0x00000004
#define SENSOR_WAIT_VSYNC        0x00000008
#ifdef CAMSENSOR_DEBUG
#define SENSOR_PREV_DEBUG        0x00000010
#endif

#define SENSOR_FLASH_OFF         0x00000000
#define SENSOR_FLASH_ON          0x00000100
#define SENSOR_PRE_FLASH_ON      0x00000200
#define SENSOR_VIDEO_LIGHT_ON    0x00000400
#define SENSOR_FLASH_WAIT_VSYNC  0x00000800

#define SENSOR_AF_OFF            0x00000000
#define SENSOR_AF_STANDBY        0x00001000
#define SENSOR_AF_SEND_CMD       0x00002000
#define SENSOR_AF_FOCUSING       0x00004000
#define SENSOR_AF_COMPLETE       0x00008000
#define SENSOR_AF_CANCEL         0x00010000
#define SENSOR_AF_WAIT_VSYNC     0x00020000

#define MODE_PREV_VREF_INT       0x00000001
#define MODE_STILL_VREF_INT      0x00000002
#define AF_VREF_INT              0x00000004
#define PRE_FLASH_ON_VREF_INT    0x00000008
#define FLASH_ON_VREF_INT        0x00000010
#define FLASH_OFF_VREF_INT       0x00000020

#define VSYNC_HIGH        1
#define VSYNC_LOW         2

#define PREV_STILL_MODE           0
#define PREV_CAMCORDER_MODE       1
#define PREV_CAMCORDER_720P_MODE  2

#define STILL_YUV_MODE        0
#define STILL_JPEG_MODE       1

#define FLASH_OFF_MODE        0
#define FLASH_STILL_MODE      1
#define FLASH_VIDEOLIGHT_MODE 2

#define AF_OFF_MODE           0
#define AF_SINGLE_MODE        1
#define AF_MACRO_MODE         2
#define AF_INFINITY_MODE      3

#define DEFAULT_STILL_5M_WIDTH  2592
#define DEFAULT_STILL_5M_HEIGHT 1944
#define DEFAULT_PREVIEW_SIZE_WIDTH  640
#define DEFAULT_PREVIEW_SIZE_HEIGHT 480

#define VIDEO_720P_WIDTH 1280
#define VIDEO_720P_HEIGHT 720

#define PREVIEW_SXGA_WIDTH 1280
#define PREVIEW_SXGA_HEIGHT 960
#define VIDEO_WVGA_WIDTH   800
#define VIDEO_WVGA_HEIGHT  480
#define VIDEO_VGA_WIDTH    640
#define VIDEO_VGA_HEIGHT   480

#define ERR_CODE -1

#define AF_WINDOW_WIDTH_SIZE_MAX   1350
#define AF_WINDOW_WIDTH_SIZE_MIN      2
#define AF_WINDOW_HEIGHT_SIZE_MAX  1350
#define AF_WINDOW_HEIGHT_SIZE_MIN     8
#define AF_WINDOW_OFFSET_X           49
#define AF_WINDOW_OFFSET_Y            4
#define AF_WINDOW_UNS_X              33
#define AF_WINDOW_UNS_Y               5
#define AF_WINDOW_WIDTH_DEF         230
#define AF_WINDOW_HEIGHT_DEF        230
#define AF_WINDOW_POSX_DEF         1230
#define AF_WINDOW_POSY_DEF          861
#define AE_WINDOW_WIDTH_DEF         288
#define AE_WINDOW_HEIGHT_DEF        277
#define AE_WINDOW_WIDTH_MAX         288
#define AE_WINDOW_WIDTH_MIN          32
#define AE_WINDOW_HEIGHT_MAX        277
#define AE_WINDOW_HEIGHT_MIN          8
#define AWB_WINDOW_WIDTH_DEF        288
#define AWB_WINDOW_HEIGHT_DEF       277
#define AWB_WINDOW_WIDTH_MAX        288
#define AWB_WINDOW_WIDTH_MIN         32
#define AWB_WINDOW_HEIGHT_MAX       277
#define AWB_WINDOW_HEIGHT_MIN         8

#define FPS_30 30
#define FPS_24 24
#define FPS_15 15
#define FPS_10 10
#define FPS_7P5 7
#define FPS_5   5

#define FLASH_THRESHOLD 0x797C

#define EXIF_FLASH_FLASH_ON         0x01
#define EXIF_FLASH_COMP_FIRE        0x08
#define EXIF_FLASH_COMP_SUPPRES     0x10
#define EXIF_FLASH_AUTO_FIRE        0x18

#define EXIF_F_NUMBER               3

#define FOCUS_MECH_INF           0

#define REG_OFFSET_2541              0x4852
#define REG_OFFSET_2367         0x486A
#define REG_OFFSET_2368    0x486C
#define REG_OFFSET_2370     0x4870
#define REG_OFFSET_2369    0x486E
#define REG_OFFSET_2371     0x4872
#define REG_OFFSET_2406   0x4822
#define REG_OFFSET_2521     0x4838
#define REG_OFFSET_2407          0x4824
#define REG_OFFSET_2373          0x4876
#define REG_OFFSET_2374         0x4878
#define REG_OFFSET_2375          0x487A
#define REG_OFFSET_2376         0x487C
#define REG_OFFSET_2377          0x487E
#define REG_OFFSET_2378         0x4880
#define REG_OFFSET_2362           0x4844

struct cam_af_offset_t {
  int offset_2541;
  int offset_2367;
  int offset_2368;
  int offset_2370;
  int offset_2369;
  int offset_2371;
  int offset_2406;
  int offset_2521;
  int offset_2407;
  int offset_2373;
  int offset_2374 ;
  int offset_2375 ;
  int offset_2376 ;
  int offset_2377 ;
  int offset_2378 ;
  int offset_2362 ;
};

static struct  cam_af_offset_t cam_preview_sxga_offset;
static struct  cam_af_offset_t cam_preview_720p_offset;
static struct  cam_af_offset_t cam_preview_wvga_offset;
static struct  cam_af_offset_t cam_preview_vga_offset;

#define OUTPUT_HALF_RELEASE_IMAGE 1

typedef enum
{
  CAMERA_INIT_STATE,
  CAMERA_POWERON_STATE,
  CAMERA_PREVIEW_STATE,
  CAMERA_STILL_STATE,
} camera_state_type;

typedef enum
{
  CAMERA_IDLE_FOCUS,
  CAMERA_AF_FOCUS,
  CAMERA_MOVE_FOCUS,
} camera_focus_state_type;

typedef enum
{
  CAMERA_FOCUS_SEARCH_FAR,
  CAMERA_FOCUS_SEARCH_MACRO,
} camera_focus_search_type;

typedef enum
{
  CAMERA_V_FALL_NONE,
  CAMERA_V_FALL_AF_MODE = 1,
  CAMERA_V_FALL_STILL,

  CAMERA_V_FALL_AF_PRE_FLASH,
  CAMERA_V_FALL_AF_PRE_FLASH_HALF_REL_CHECK,
  CAMERA_V_FALL_AF_PRE_FLASH_AE_AWB_STS_CHECK,
  CAMERA_V_FALL_AF_PRE_FLASH_AF_STS_CHECK,
  CAMERA_V_FALL_YUV_STILL_MAIN_FLASH_OFF_AF_OFF,
  CAMERA_V_FALL_YUV_STILL_MAIN_FLASH_OFF_AF_ON,
  CAMERA_V_FALL_JPEG_STILL_MAIN_FLASH_OFF_AF_OFF,
  CAMERA_V_FALL_JPEG_STILL_NORMAL,

  CAMERA_V_FALL_JPEG_STILL_MAIN_FLASH_ON_AF_ON,

  CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE,
  CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW,

  CAMERA_V_FALL_DEFAULT_FOCUS,
  CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK,
  CAMERA_V_FALL_DEFAULT_FOCUS_MOVE,

} camera_vsync_fall_irq_type;

typedef enum
{
    CAMERA_V_RISE_NONE,
    CAMERA_V_RISE_SCENE_MODE = 1,

    CAMERA_V_RISE_YUV_STILL_MAIN_FLASH_AF_OFF,
    CAMERA_V_RISE_YUV_STILL_MAIN_FLASH_AF_ON,

    CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_RESUME_AF_ON,
    CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_ON_AF_OFF,
    CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_HALFREL_CHANGE_CHECK_AF_OFF,
    CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_AE_AWB_STS_CHECK_AF_OFF,
    CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_MF_OTICAL_INF_CHECK_AF_OFF,
    CAMERA_V_RISE_YUV_JPEG_STILL_MAIN_FLASH_AF_OFF,

    CAMERA_V_RISE_JPEG_STILL_MAIN_FLASH_AF_ON,
    CAMERA_V_RISE_JPEG_STILL_NORMAL,
    CAMERA_V_RISE_MAIN_FLASH_GET_WAIT_TIME_WAIT,
    CAMERA_V_RISE_MAIN_FLASH_GET_WAIT_TIME,

} camera_vsync_rise_irq_type;

typedef enum
{
    CAMERA_POWER_ON_NORMAL,
    CAMERA_POWER_ON_MOVIE_720P,
    CAMERA_POWER_ON_MOVIE_WVGA,
    CAMERA_POWER_ON_MOVIE_VGA,
    CAMERA_POWER_ON_TEST_DL,
} camera_power_on_reg_type;

enum cam_af_cmd_id {
  IDLE_CMD,
  ABORT_CMD,
  SLEEP_CMD,
  INIT_CMD,
  MANUAL_CMD,
  SINGLE_CMD,
  CONTINUOUS_CMD,
  MAX_AF_CMD,
};

struct cam_af_cmd {
  enum cam_af_cmd_id cmd;
  uint16_t param;
};

#define PRE_FLASH_CURRENT   (unsigned short)200
#define FLASH_CURRENT       (unsigned short)200
#define VIDEO_LIGHT_CURRENT (unsigned short)25

struct cam_work_t {
    struct work_struct work;
};

struct cam_ctrl_t {

  const struct msm_camera_sensor_info *sensordata;

  int wait_flag;

  struct cam_reg_t *cam_regs;
  unsigned int hw_ver;

  unsigned int sensor_state;

  uint8_t  preview_mode;
  uint8_t  still_mode;
  uint8_t  flash_mode;
  uint8_t  af_mode;
  uint8_t  brightness_mode;
  uint8_t  contrast_mode;
  uint8_t  sharpness_mode;
  uint8_t  scene_mode;
  uint8_t  effect_mode;
  uint8_t  wb_mode;

  uint16_t current_picture_width;
  uint16_t current_picture_height;

  uint16_t current_display_width;
  uint16_t current_display_height;
  uint16_t past_picture_width;
  uint16_t past_picture_height;
  uint16_t past_display_width;
  uint16_t past_display_height;

  int picture_size_chenge;
  camera_state_type camera_state;
  camera_focus_state_type focus_state;
  camera_focus_search_type focus_search_state;
  uint8_t  af_result;

  camera_vsync_fall_irq_type  v_fall_proc;
  camera_vsync_rise_irq_type  v_rise_proc;

  uint8_t current_effect;
  uint8_t current_wb;
  uint8_t current_iso;
  uint8_t current_ae_mode;

  uint8_t current_wb_update;
  uint8_t current_iso_update;
  uint8_t current_ae_mode_update;

  int still_err_flg;

  struct afwin_info_cfg current_afwin_info;
  struct aeawb_win_info_cfg current_aewin_info;
  struct aeawb_win_info_cfg current_awbwin_info;

  unsigned short current_fps;
  unsigned short past_fps;

  unsigned char current_lens_shading;

  camera_scene_type current_scene;

  struct cam_i2c_reg_conf* set_scene_reg;
  uint16_t set_scene_reg_size;
  int scene_err_flg;

  uint8_t current_scene_update;

  unsigned int vref_high_int_func;
  unsigned int vref_low_int_func;

  uint16_t vsync_count;
  struct cam_af_cmd af_cmd;
  struct sensor_pict_fps gfps;

  unsigned long cam006_o0_H;
  unsigned long cam006_o0_M;
  unsigned long cam006_o0_L;
  unsigned long cam006_o1_H;
  unsigned long cam006_o1_M;
  unsigned long cam006_o1_L;

  camera_flash_parm_type flash_parm_info;
  uint16_t main_flash_cg_offset;
  uint16_t main_flash_sc_shift_r;
  uint16_t main_flash_sc_shift_b;
  int af_proc_call_flg;
  int main_flash_on_flg;

  int pre_flash_on_flg;

  uint32_t main_flash_off_v_cnt;
  uint32_t yuv_main_flash_on_v_cnt;

  camera_set_af_otp_data_type cam006_af_otp_data;

  uint32_t normal_still_v_cnt;

#ifdef TIME_CHECK
  struct timeval vsync_high_tv;
  struct timeval prev_vsync_high_tv;

  struct timeval vsync_low_tv;
  struct timeval prev_vsync_low_tv;

  struct timeval seq_tv;
  struct timeval prev_seq_tv;
#endif

  int flash_on_flg;

  int internal_power_offon_retry_flg;

  uint32_t  exif_exposure_time;

  int aw_watch_cnt;
  camera_power_on_reg_type  power_on_type;

};

static struct  cam_work_t *cam_sensorw;
static struct  i2c_client *cam_client;
static struct  cam_ctrl_t *cam_ctrl;

static DECLARE_WAIT_QUEUE_HEAD(cam_wait_queue);
DECLARE_MUTEX(cam_sem);

#define SCENE_COMMON_BRIGHTNESS 4
#define SCENE_COMMON_EV 0

static camera_scene_info_type scene_info[ CAM_SCENE_MAX ] =
{

  {
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_CENTER,
    CAM_ISO_AUTO,
    CAM_FLASH_AUTO,
    SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_AUTO,
    SCENE_COMMON_EV,
    FPS_15,
    3,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },

  {
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_CENTER,
    CAM_ISO_AUTO,
    CAM_FLASH_AUTO,
    SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_AUTO,
    SCENE_COMMON_EV,
    FPS_15,
    2,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },

  {
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_AVERAGE,
    CAM_ISO_AUTO,
    CAM_FLASH_AUTO,
    SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_AUTO,
    SCENE_COMMON_EV,
    FPS_15,
    4,
    4,
    4,
    CAM_EFFECT_NONE,
    TRUE
  },

  {
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_CENTER,
    CAM_ISO_AUTO,
    CAM_FLASH_ON,
    SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_AUTO,
    SCENE_COMMON_EV,
    FPS_15,
    2,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },

  {
    CAM_WB_AUTO,
    CAM_AF_MODE_NORMAL,
    CAM_AEC_AVERAGE,
    CAM_ISO_AUTO,
    CAM_FLASH_OFF,
    SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_AUTO,
    SCENE_COMMON_EV,
    FPS_15,
    3,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },

  {
    CAM_WB_AUTO,
    CAM_AF_MODE_NORMAL,
    CAM_AEC_CENTER,
    CAM_ISO_400,
    CAM_FLASH_OFF,
    SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_AUTO,
    SCENE_COMMON_EV,
    FPS_15,
    3,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },
};

#define V_FALL_PROC_RESERVE_MAX 10
static camera_vsync_fall_irq_type cam006_v_fall_proc_reserve_tbl[V_FALL_PROC_RESERVE_MAX] =
{
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE,
    CAMERA_V_FALL_NONE
};

static int32_t cam_i2c_write(unsigned short saddr,
    unsigned short waddr, unsigned long wdata, enum cam_width_t width);

static int32_t cam_i2c_write_table(
    struct cam_i2c_reg_conf const *reg_conf_tbl,
    int num_of_items_in_table);

static int32_t cam_i2c_read(unsigned short   saddr,
    unsigned short raddr, unsigned short *rdata, enum cam_width_t width);

static int32_t cam_i2c_read_long(unsigned short   saddr,
    unsigned short raddr, unsigned long *rdata, enum cam_width_t width);

int cam006_check_internal_status( unsigned short comp_stat_data, unsigned char clr_data );
static int L_cam006_kernel_sensor_power_off( void );

#ifndef CONFIG_TARGET_PROTOTYPE_WS1

static int32_t cam_i2c_flash_write(unsigned short saddr,
    unsigned char data1, unsigned char data2, unsigned char data3 );

#endif

int cam006_initial_setting_1( camera_power_on_reg_type power_on_type );
int cam006_o_data_read( void );
int cam006_check_internal_status( unsigned short comp_stat_data, unsigned char clr_data );
int cam006_initial_setting_2( camera_power_on_reg_type power_on_type );
int cam006_initial_setting_3( void );
int cam006_initial_write_rom_correction( void );
int cam006_initial_setting_4( camera_power_on_reg_type power_on_type );
int cam006_AFdriver_data_write( void );
int cam006_initial_setting_5( camera_power_on_reg_type power_on_type );
int cam006_initial_all_reg_read( void );
static int L_cam006_start_preview( void );
static int L_cam006_kernel_sensor_power_on( camera_power_on_reg_type power_on_type );
static int L_cam006_kernel_sensor_dlc_power_on( void );

static int L_cam006_kernel_move_focus_infinity( void );

static int L_cam006_kernel_move_focus_optical_infinity(void);

static void L_cam006_entry_vsync_low_proc( camera_vsync_fall_irq_type entry );
static void L_cam006_entry_vsync_high_proc( camera_vsync_rise_irq_type entry );

static void L_cam006_delete_vsync_high( void );

static int L_cam006_kernel_pre_flash( void );
static int L_cam006_kernel_main_flash( void );
static int L_cam006_kernel_flash_off( void );
static int L_cam006_kernel_flash_seq_auto_focus_ready( void );
static int L_cam006_kernel_still_flash_end( void );

static int L_cam006_kernel_flash_threshold_check( int* flash_flg );

static int L_cam006_kernel_set_main_flash_offset( void );

static int L_cam006_kernel_still_movie_chg_retry_set_image_control(camera_power_on_reg_type);

static int L_cam006_kernel_move_focus( uint16_t focus_pos );
static int L_cam006_kernel_flash_seq_auto_focus_off_ready( void );

static int L_cam006_kernel_check_flash_on( int* flash_onoff_flg );

#define DELAY_MARGIN 63
static int L_cam006_kernel_calc_capture_wait_time( int32_t* wait_time );

static int L_cam006_kernel_get_shutter_speed(void);

static int L_cam006_kernel_get_now_sensor_mode( uint16_t* current_mode );
static int L_cam006_kernel_set_v_fall_proc( camera_vsync_fall_irq_type v_fall_proc );
static int L_cam006_kernel_get_v_fall_proc( camera_vsync_fall_irq_type* v_fall_proc_reserve );
static int L_cam006_kernel_get_vfall_proc_table( camera_vsync_fall_irq_type* v_fall_proc_reserve );
static void L_cam006_kernel_set_vfall_free_irq( void );

#ifdef CAMSENSOR_DEBUG
static int cam_dbg_prev_mode       = 0;
static int cam_dbg_preflash_mode   = 0;
static int cam_dbg_flash_mode      = 0;
static int cam_dbg_videolight_mode = 0;

#define PREVIEW_1_NORMAL 0
#define PREVIEW_1_NIGHT  1
#define PREVIEW_2_NORMAL 2
#define PREVIEW_2_NIGHT  3
#define PREVIEW_2_FIXED  4
#define PREVIEW_3_NORMAL 5
#define PREVIEW_3_NIGHT  6
#define PREVIEW_3_FIXED  7

static void cam_set_still_debug_mode( void )
{
  

  switch(cam_dbg_prev_mode)
  {
    case PREVIEW_1_NIGHT:
    case PREVIEW_2_NIGHT:
    case PREVIEW_3_NIGHT:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->still_night_reg_settings[0],
        cam_ctrl->cam_regs->still_night_reg_settings_size);
      break;
    default:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->still_normal_reg_settings[0],
        cam_ctrl->cam_regs->still_normal_reg_settings_size);
      break;
  }
}

static void cam_set_prev_debug_mode( void )
{
  

  switch(cam_dbg_prev_mode)
  {
    case PREVIEW_1_NORMAL:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview1_normal_reg_settings[0],
       cam_ctrl->cam_regs->preview1_normal_reg_settings_size);
      break;
    case PREVIEW_1_NIGHT:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview1_night_reg_settings[0],
       cam_ctrl->cam_regs->preview1_night_reg_settings_size);
      break;
    case PREVIEW_2_NORMAL:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview2_normal_reg_settings[0],
       cam_ctrl->cam_regs->preview2_normal_reg_settings_size);
      break;
    case PREVIEW_2_NIGHT:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview2_night_reg_settings[0],
        cam_ctrl->cam_regs->preview2_night_reg_settings_size);
      break;
    case PREVIEW_2_FIXED:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview2_fix_reg_settings[0],
        cam_ctrl->cam_regs->preview2_fix_reg_settings_size);
      break;
    case PREVIEW_3_NORMAL:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview3_normal_reg_settings[0],
        cam_ctrl->cam_regs->preview3_normal_reg_settings_size);
      break;
    case PREVIEW_3_NIGHT:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview3_night_reg_settings[0],
        cam_ctrl->cam_regs->preview3_night_reg_settings_size);
      break;
    case PREVIEW_3_FIXED:
      
      cam_i2c_write_table(&cam_ctrl->cam_regs->preview3_fix_reg_settings[0],
        cam_ctrl->cam_regs->preview3_fix_reg_settings_size);
      break;
    default:
      
      break;
  }
}

static int cam628_flash_disable(void);
static unsigned int cam_get_state(unsigned int type);
static unsigned int cam_chg_state(unsigned int type, unsigned int state);

void cam_set_flash_debug_mode(int mode)
{

  cam_ctrl->flash_mode = mode;

  if(cam_ctrl->flash_mode == FLASH_STILL_MODE) {
    unsigned int flash_state = 0x00000000;

    flash_state = cam_get_state(SENSOR_STATE_FLASH);
    if(flash_state & SENSOR_VIDEO_LIGHT_ON )
    {
      cam_chg_state(SENSOR_STATE_FLASH, SENSOR_FLASH_OFF);
      cam628_flash_disable();
    }
  }

  
}

#endif

void cam_set_flash_debug_mode(int mode)
{

  cam_ctrl->flash_mode = mode;

}

#ifdef AF_FLASH_BB_CHECK
#define GPIO_LO (0)
#define GPIO_HI (1)

int cam_af_flash_bb_check(unsigned mode, unsigned on_off)
{
  int rc = 0;

  switch(mode) {
    case BB_CHK_AF:
      if(on_off) {

        gpio_direction_output( AF_CHK_GPIO, GPIO_HI );
      }
      else {

        gpio_direction_output( AF_CHK_GPIO, GPIO_LO );
      }
      break;
    case BB_CHK_STILL:
      if(on_off) {

        gpio_direction_output( FL1_IN_GPIO, GPIO_HI );
      }
      else {

        gpio_direction_output( FL1_IN_GPIO, GPIO_LO );
      }
      break;
    default:

      rc = -1;
      break;
  }

  return rc;
}
#endif

#ifdef _FLASH_ENABLE_

struct cam628_work_t {
    struct work_struct work;
};

struct cam628_i2c_reg_conf {
  unsigned short mA;
    unsigned char  data;
};

struct cam628_i2c_reg_conf flash_reg_tbl[] =
{
  {  200, 0x18 },
  {  300, 0x5B },
  {  325, 0x5D },
  {  337, 0xD5 },
  {  350, 0x1E },
  {  375, 0x9A },
  {  400, 0x13 },
  {  412, 0xD9 },
  {  450, 0x15 },
  {  487, 0xDD },
  {  500, 0x17 },
  {  525, 0xDF },
  {  550, 0x19 },
  {  600, 0x1B },
  {  650, 0x1D },
  {  675, 0x95 },
  {  700, 0x1F },
  {  750, 0x97 },
  {  825, 0x99 },
  {  875, 0x9B },
  {  950, 0x9D },
  { 1000, 0x9F },
};

struct cam628_i2c_reg_conf spotlight_reg_tbl[] =
{
  {  25, 0x42 },
  {  37, 0xC2 },
  {  50, 0x02 },
  {  75, 0x46 },
  { 100, 0x04 },
  { 112, 0xC6 },
  { 125, 0x4A },
  { 150, 0x06 },
  { 175, 0x4E },
  { 187, 0xCA },
  { 200, 0x08 },
  { 225, 0x45 },
  { 250, 0x47 },
  { 262, 0xCE },
  { 275, 0x49 },
};

static struct  cam628_work_t *cam628_sensorw;
static struct  i2c_client *cam628_client;

static const struct msm_camera_sensor_info *flash_data;

static DECLARE_WAIT_QUEUE_HEAD(cam628_wait_queue);

#define FLASH      0
#define SPOT_LIGHT 1

#define ADDR                      0x02
#define FLASH_OFF_REG             0x00

#define OVERSHOOT_FIX_CURRENT     400

static int cam628_init_client(struct i2c_client *client)
{

    init_waitqueue_head(&cam628_wait_queue);
    return 0;
}

static int cam628_i2c_probe(struct i2c_client *client,
    const struct i2c_device_id *id)
{
    int rc = 0;
    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        rc = -ENOTSUPP;
        goto probe_failure;
    }

    cam628_sensorw =
        kzalloc(sizeof(struct cam628_work_t), GFP_KERNEL);

    if (!cam628_sensorw) {
        rc = -ENOMEM;
        goto probe_failure;
    }

    i2c_set_clientdata(client, cam628_sensorw);
    cam628_init_client(client);
    cam628_client = client;

    return 0;

probe_failure:
    kfree(cam628_sensorw);
    cam628_sensorw = NULL;
    
    return rc;
}

static int __exit cam628_remove(struct i2c_client *client)
{
    struct cam628_work_t *sensorw = i2c_get_clientdata(client);
    free_irq(client->irq, sensorw);
    cam628_client = NULL;
    cam628_sensorw = NULL;
    kfree(sensorw);
    return 0;
}

static const struct i2c_device_id cam628_i2c_id[] = {
    { "camera_flash_drv", 2},
    { },
};

static struct i2c_driver cam628_i2c_driver = {
    .id_table = cam628_i2c_id,
    .probe  = cam628_i2c_probe,
    .remove = __exit_p(cam_remove),
    .driver = {
        .name = "camera_flash_drv",
    },
};

static int cam628_init_probe( void )
{
  int rc = i2c_add_driver(&cam628_i2c_driver);
    if (rc < 0) {
      
        rc = -ENOTSUPP;
        goto probe_fail;
    }

    
    return rc;

probe_fail:

    return rc;
}

static void cam628_flash_en_off( void )
{
  

  if( flash_data && flash_data->flash_config) {
    flash_data->flash_config(flash_data->flash_en, 0);
  }
  else {

  }
}

static void cam628_flash_set_off( void )
{

  if( flash_data && flash_data->flash_config) {
    flash_data->flash_config(flash_data->flash_set, 0);
  }
  else {

  }
}

static int cam628_flash_disable(void)
{
  int rc = 0;

  

  cam628_flash_set_off();

  mdelay(1);

  cam628_flash_en_off();

  return rc;
}

static void cam628_set_data(const struct msm_camera_sensor_info *data)
{
  flash_data = data;
}

#endif

#ifdef SENSOR_ERROR_CHK_LOG
static void sensor_error_check( void )
{

  int reg_rc = 0;
  unsigned short read_value = 0x0000;

  reg_rc = cam_i2c_write(cam_client->addr, 0x002C, 0x7000, WORD_LEN);
  reg_rc = cam_i2c_write(cam_client->addr, 0x002E, 0x026E, WORD_LEN);
  reg_rc = cam_i2c_read(cam_client->addr,  0x0F12, &read_value, WORD_LEN);

  if(reg_rc < 0) MSG_FATAL("*** ErrorInfo i2c error!\n",0,0,0);
  MSG_FATAL("*** ErrorInfo Chk 0x%04x!\n",read_value,0,0);

  reg_rc = cam_i2c_write(cam_client->addr, 0x002C, 0x7000, WORD_LEN);
  reg_rc = cam_i2c_write(cam_client->addr, 0x002E, 0x0250, WORD_LEN);
  reg_rc = cam_i2c_read(cam_client->addr,  0x0F12, &read_value, WORD_LEN);

  if(reg_rc < 0) MSG_FATAL("*** InitHwErr i2c error!\n",0,0,0);
  MSG_FATAL("*** InitHwErr Chk 0x%04x!\n",read_value,0,0);

  reg_rc = cam_i2c_write(cam_client->addr, 0x002C, 0x7000, WORD_LEN);
  reg_rc = cam_i2c_write(cam_client->addr, 0x002E, 0x02A8, WORD_LEN);
  reg_rc = cam_i2c_read(cam_client->addr,  0x0F12, &read_value, WORD_LEN);

  if(reg_rc < 0) MSG_FATAL("*** ErrorPrevConfig i2c error!\n",0,0,0);
  MSG_FATAL("*** ErrorPrevConfig Chk 0x%04x!\n",read_value,0,0);

  reg_rc = cam_i2c_write(cam_client->addr, 0x002C, 0x7000, WORD_LEN);
  reg_rc = cam_i2c_write(cam_client->addr, 0x002E, 0x02C0, WORD_LEN);
  reg_rc = cam_i2c_read(cam_client->addr,  0x0F12, &read_value, WORD_LEN);

  if(reg_rc < 0) MSG_FATAL("*** AfCmdError i2c error!\n",0,0,0);
  MSG_FATAL("*** AfCmdError Chk 0x%04x!\n",read_value,0,0);

}
#endif

#ifdef TIME_CHECK
static void get_vsync_time(struct timeval *tv, struct timeval *prev_tv)
{
  long sec = 0;
  long usec = 0;
  long diff_sec = 0;
  long diff_usec = 0;

  do_gettimeofday(tv);

  if( prev_tv->tv_usec > tv->tv_usec )
  {
    tv->tv_usec += 1000000;
    tv->tv_sec--;
  }

  diff_sec = tv->tv_sec - prev_tv->tv_sec;
  diff_usec = tv->tv_usec - prev_tv->tv_usec;

  if(diff_sec != 0) {
    sec = diff_sec * 1000;
  }

  if( diff_usec != 0 ) {
    usec = diff_usec / 1000;
  }

  

  prev_tv->tv_sec = tv->tv_sec;
  prev_tv->tv_usec = tv->tv_usec;
}
#endif

static unsigned int cam_chg_state(unsigned int type, unsigned int state)
{

  cam_ctrl->sensor_state &= ~type;
  cam_ctrl->sensor_state |= state;

  return cam_ctrl->sensor_state;
}

static unsigned int cam_get_state(unsigned int type)
{
  return cam_ctrl->sensor_state & type;
}

static int cam_wait_event( void )
{
  

  return wait_event_interruptible(cam_wait_queue,cam_ctrl->wait_flag != 0);
}

static void cam_wake_up( void )
{
  

  cam_ctrl->wait_flag = 1;
  wake_up_interruptible(&cam_wait_queue);
}

static irqreturn_t vsync_high_interrupt(int irq, void *handle);
static irqreturn_t vsync_low_interrupt(int irq, void *handle);

static void cam_tasklet_proc_high( unsigned long data )
{
  int rc = TRUE;

  uint16_t read_data;
  int32_t wait_time;

  if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_SCENE_MODE )
  {

      rc = cam_i2c_write_table((struct cam_i2c_reg_conf const *)cam_ctrl->set_scene_reg,
                                   cam_ctrl->set_scene_reg_size );
      if( rc < 0 )
      {
          ERR_LOG("VSYNC HIGH CAMERA_V_RISE_SCENE_MODE I2C Err\n");
          cam_ctrl->scene_err_flg = TRUE;
          goto v_h_err_end_thread;
      }

      L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

      cam_wake_up();
  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_ON_AF_OFF )
  {

      rc = L_cam006_kernel_flash_seq_auto_focus_off_ready();
      if( rc < 0 )
      {
          ERR_LOG("L_cam006_kernel_flash_seq_auto_focus_off_ready() Err\n");
          goto v_h_err_end_thread;
      }

      rc = cam_i2c_read( cam_client->addr, 0x0004, &read_data, ADWORD_DATBYTE_LEN );
      if( rc < 0 )
      {
          ERR_LOG("CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_ON_AF_OFF 0x0004 Read Err\n");
          goto v_h_err_end_thread;
      }

      if(( read_data & 0x03 ) == OUTPUT_HALF_RELEASE_IMAGE )
      {

          L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_AE_AWB_STS_CHECK_AF_OFF );
      }
      else
      {

          L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_HALFREL_CHANGE_CHECK_AF_OFF );
      }

  }
  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_HALFREL_CHANGE_CHECK_AF_OFF )
  {

      rc = cam_i2c_read( cam_client->addr, 0x00F8, &read_data, ADWORD_DATBYTE_LEN );
      if( rc < 0 )
      {
          ERR_LOG("Vsync Fall AF FLASH HalfRel Change Err\n");
          goto v_h_err_end_thread;
      }

      if( read_data & 0x02 )
      {

          L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_AE_AWB_STS_CHECK_AF_OFF );
      }
  }
  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_AE_AWB_STS_CHECK_AF_OFF )
  {

      rc = cam_i2c_read( cam_client->addr, 0x02AA, &read_data, ADWORD_DATBYTE_LEN );
      if( rc < 0 )
      {
          ERR_LOG("Vsync Rise AF OFF PreFlash AE AWB STS Check Err\n");
          goto v_h_err_end_thread;
      }

      if( read_data != 0x00 )
      {

          return ;
      }

      rc = cam_i2c_read( cam_client->addr, 0x0286, (uint16_t*)&cam_ctrl->flash_parm_info.pre_after_es, WORD_LEN );
      if( rc < 0 )
      {
          ERR_LOG("VSYNC HIGH PRE FLASH After 0x0286 Read Err\n");
          goto v_h_err_end_thread;
      }
      rc = cam_i2c_read( cam_client->addr, 0x028A, &cam_ctrl->flash_parm_info.pre_after_as, WORD_LEN );
      if( rc < 0 )
      {
          ERR_LOG("VSYNC HIGH PRE FLASH After 0x028A Read Err\n");
          goto v_h_err_end_thread;
      }

      rc = cam_i2c_read( cam_client->addr, 0x026E, &cam_ctrl->flash_parm_info.pre_after_rt_r, WORD_LEN );
      if( rc < 0 )
      {
          ERR_LOG("VSYNC HIGH PRE FLASH After 0x026E Read Err\n");
          goto v_h_err_end_thread;
      }
      rc = cam_i2c_read( cam_client->addr, 0x0270, &cam_ctrl->flash_parm_info.pre_after_rt_b, WORD_LEN );
      if( rc < 0 )
      {
          ERR_LOG("VSYNC HIGH PRE FLASH After 0x0270 Read Err\n");
          goto v_h_err_end_thread;
      }

      rc = L_cam006_kernel_move_focus( cam_ctrl->cam006_af_otp_data.focus_optical_inf );
      if( rc < 0 )
      {
          ERR_LOG("VSYNC_HIGH_PRE_FLASH After Move Focus Optical Inf Err\n");
          goto v_h_err_end_thread;
      }

      L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_MF_OTICAL_INF_CHECK_AF_OFF );
  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_MF_OTICAL_INF_CHECK_AF_OFF )
  {

      rc = cam_i2c_read( cam_client->addr, 0x00F8, &read_data, ADWORD_DATBYTE_LEN );
      if( rc < 0 )
      {
          ERR_LOG( "CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_MF_OTICAL_INF_CHECK_AF_OFF mode change Err\n");
          goto v_h_err_end_thread;
      }

      if( read_data & 0x10 )
      {

          rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x16, ADWORD_DATBYTE_LEN );

          if( rc < 0 )
          {
              ERR_LOG("optical Inf Move Lock Clear Err\n");
              goto v_h_err_end_irq;
          }

          L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

          cam_wake_up();
      }
  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_JPEG_STILL_MAIN_FLASH_AF_OFF )
  {

      rc = L_cam006_kernel_calc_capture_wait_time( &wait_time );
      if( rc < 0 )
      {
          ERR_LOG("Cap Wait Time Err\n");
          goto v_h_err_end_irq;
      }

      if( wait_time != 0 )
      {
          mdelay( wait_time );
      }

      cam_wake_up();

      rc = L_cam006_kernel_set_main_flash_offset();
      if( rc < 0 )
      {
          ERR_LOG("L_cam006_kernel_set_main_flash_offset Err\n");
          goto v_h_err_end_irq;
      }

      if( cam_ctrl->still_mode == STILL_YUV_MODE )
      {

          free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

          L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_YUV_STILL_MAIN_FLASH_OFF_AF_OFF );
          L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

          if (request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0))
          {
              ERR_LOG("SENSOR_SNAPSHOT_MODE Vsycn Fall Trig Set Err\n");
              goto v_h_err_end_irq;
          }
      }
      else if( cam_ctrl->still_mode == STILL_JPEG_MODE )
      {

          rc = cam_i2c_write( cam_client->addr, 0x0011, 0x02, ADWORD_DATBYTE_LEN );
          if( rc < 0 )
          {
              ERR_LOG("Vsync Rise Capture Mode Change Err\n");
              goto v_h_err_end_irq;
          }

          free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

          L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_JPEG_STILL_MAIN_FLASH_OFF_AF_OFF );

          L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

          if (request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0))
          {
              ERR_LOG("SENSOR_SNAPSHOT_MODE Vsycn Fall Trig Set Err\n");
              goto v_h_err_end_irq;
          }
      }
  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_STILL_MAIN_FLASH_AF_ON )
  {

      cam_wake_up();

      rc = L_cam006_kernel_set_main_flash_offset();
      if( rc < 0 )
      {
          ERR_LOG("Vsync Rise Main Flash Offset Err\n");
          goto v_h_err_end_irq;
      }

      L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_RESUME_AF_ON );

  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_RESUME_AF_ON )
  {

      rc = L_cam006_kernel_pre_flash();
      if( rc < 0 )
      {
          ERR_LOG("SENSOR_SNAPSHOT_MODE L_cam006_kernel_pre_flash\n");
          goto v_h_err_end_irq;
      }

      cam_ctrl->pre_flash_on_flg = TRUE;

      free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

      if( cam_ctrl->still_mode == STILL_YUV_MODE )
      {
          L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_YUV_STILL_MAIN_FLASH_OFF_AF_ON );
      }
      else if( cam_ctrl->still_mode == STILL_JPEG_MODE )
      {
          L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_JPEG_STILL_MAIN_FLASH_ON_AF_ON );
      }

      L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

      if (request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0))
      {
          ERR_LOG("SENSOR_SNAPSHOT_MODE Vsycn Fall Trig Set Err\n");
          cam_ctrl->still_err_flg = TRUE;
          goto v_h_err_end_irq;
      }
  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_JPEG_STILL_NORMAL )
  {

      rc = L_cam006_kernel_calc_capture_wait_time( &wait_time );
      if( rc < 0 )
      {
          ERR_LOG("Cap Wait Time Err\n");
          goto v_h_err_end_irq;
      }

      if( wait_time != 0 )
      {
          mdelay( wait_time );
      }

      cam_wake_up();

      rc = cam_i2c_write( cam_client->addr, 0x0011, 0x02, ADWORD_DATBYTE_LEN );
      if( rc < 0 )
      {
          ERR_LOG("Vsync Fall Capture Mode Change Err\n");
          goto v_h_err_end_irq;
      }

      free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

      L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_JPEG_STILL_NORMAL );

      L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

      if (request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0))
      {
          ERR_LOG("SENSOR_SNAPSHOT_MODE Vsycn Fall Trig Set Err\n");
          cam_ctrl->still_err_flg = TRUE;
          return;
      }
  }

  else if( cam_ctrl->v_rise_proc == CAMERA_V_RISE_JPEG_STILL_MAIN_FLASH_AF_ON )
  {

      rc = L_cam006_kernel_calc_capture_wait_time( &wait_time );
      if( rc < 0 )
      {
          ERR_LOG("Cap Wait Time Err\n");
          goto v_h_err_end_thread;
      }

      if( wait_time != 0 )
      {
          mdelay( wait_time );
      }

      cam_wake_up();

      rc = L_cam006_kernel_set_main_flash_offset();
      if( rc < 0 )
      {
          ERR_LOG("Vsync Rise Main Flash Offset Err\n");
          goto v_h_err_end_irq;
      }

      rc = cam_i2c_write( cam_client->addr, 0x0011, 0x02, ADWORD_DATBYTE_LEN );
      if( rc < 0 )
      {
          ERR_LOG("Vsync Fall Capture Mode Change Err\n");
          goto v_h_err_end_irq;
      }

      L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_RESUME_AF_ON );

  }
  return;

v_h_err_end_thread:

  cam_wake_up();

v_h_err_end_irq:
  cam_ctrl->still_err_flg = TRUE;
  L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

  return;
}

DECLARE_TASKLET( cam_tasklet_vsync_high, cam_tasklet_proc_high, ( unsigned long )NULL );

static void cam_tasklet_proc_low( unsigned long data )
{
    int rc = 0;
    unsigned short read_data;

    uint32_t flash_on_v_check_cnt = 0;
    uint32_t flash_off_v_check_cnt = 0;
    uint32_t normal_v_check_cnt = 0;

    
    #ifdef TIME_CHECK
    get_vsync_time(&cam_ctrl->vsync_low_tv, &cam_ctrl->prev_vsync_low_tv);
    #endif

    if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_AF_MODE )
    {

        rc = cam_i2c_read( cam_client->addr, 0x6D76, &read_data, ADWORD_DATBYTE_LEN );
        if ( rc < 0 )
        {
            ERR_LOG("AF Get Status Err\n");
            cam_ctrl->af_result = FALSE;
            goto err_end;
        }

        if( read_data & 0x03 )
        {

            return;
        }

        if(( read_data & 0x08 ) == 0 )
        {

            return;
        }

        rc = cam_i2c_read( cam_client->addr, 0x6D77, &read_data, ADWORD_DATBYTE_LEN );
        if ( rc < 0 )
        {
            ERR_LOG("AF Get Status Err\n");
            cam_ctrl->af_result = FALSE;
            goto err_end;
        }

        if( read_data == 0x02 )
        {

            return;
        }
        else
        {

            if( read_data == 0x00 )
            {
                cam_ctrl->af_result = FALSE;
            }
            else if( read_data == 0x01 )
            {
                cam_ctrl->af_result = TRUE;
            }

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_NONE );

            cam_wake_up();
        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_AF_PRE_FLASH )
    {

        rc = L_cam006_kernel_flash_seq_auto_focus_ready();
        if( rc < 0 )
        {
            ERR_LOG("Vsync Fall L_cam006_kernel_flash_seq_auto_focus_ready() Err\n");
            goto v_l_pre_flash_af_err_end;
        }

        rc = cam_i2c_read( cam_client->addr, 0x0004, &read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            ERR_LOG("Now Sensor Mode Read Err\n");
            goto v_l_pre_flash_af_err_end;
        }

        if(( read_data & 0x03 ) == OUTPUT_HALF_RELEASE_IMAGE )
        {

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_AF_PRE_FLASH_AE_AWB_STS_CHECK );
        }
        else
        {

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_AF_PRE_FLASH_HALF_REL_CHECK );
        }

    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_AF_PRE_FLASH_HALF_REL_CHECK )
    {

        rc = cam_i2c_read( cam_client->addr, 0x00F8, &read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            ERR_LOG("Vsync Fall AF FLASH HalfRel Change Err\n");
            goto v_l_pre_flash_af_err_end;
        }

        if( read_data & 0x02 )
        {

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_AF_PRE_FLASH_AE_AWB_STS_CHECK );
        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_AF_PRE_FLASH_AE_AWB_STS_CHECK )
    {

        rc = cam_i2c_read( cam_client->addr, 0x02AA, &read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            ERR_LOG("VSYNC LOW CAMERA_V_FALL_AF_PRE_FLASH_AE_AWB_STS_CHECK Err\n");
            goto v_l_pre_flash_af_err_end;
        }

        if( read_data != 0x00 )
        {

            return;
        }

        rc = cam_i2c_read( cam_client->addr, 0x0286, (uint16_t*)&cam_ctrl->flash_parm_info.pre_after_es, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG("VSYNC LOW PRE FLASH After 0x0286 Read Err\n");
            goto v_l_pre_flash_af_err_end;
        }
        rc = cam_i2c_read( cam_client->addr, 0x028A, &cam_ctrl->flash_parm_info.pre_after_as, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG("VSYNC LOW PRE FLASH After 0x028A Read Err\n");
            goto v_l_pre_flash_af_err_end;
        }

        rc = cam_i2c_read( cam_client->addr, 0x026E, &cam_ctrl->flash_parm_info.pre_after_rt_r, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG("VSYNC LOW PRE FLASH After 0x026E Read Err\n");
            goto v_l_pre_flash_af_err_end;
        }
        rc = cam_i2c_read( cam_client->addr, 0x0270, &cam_ctrl->flash_parm_info.pre_after_rt_b, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG("VSYNC LOW PRE FLASH After 0x0270 Read Err\n");
            goto v_l_pre_flash_af_err_end;
        }

        L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_AF_PRE_FLASH_AF_STS_CHECK );

    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_AF_PRE_FLASH_AF_STS_CHECK )
    {

        rc = cam_i2c_read( cam_client->addr, 0x6D77, &read_data, ADWORD_DATBYTE_LEN );
        if ( rc < 0 )
        {
            ERR_LOG("AF Get Status Err\n");
            cam_ctrl->af_result = FALSE;
            goto v_l_pre_flash_af_err_end;
        }

        if( read_data == 0x02 )
        {

            return;
        }
        else
        {

            rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x16, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG("AF Judge Check OK But Interrupt Flg Clear Err\n");
                goto v_l_pre_flash_af_err_end;
            }

            if( read_data == 0x00 )
            {
              cam_ctrl->af_result = FALSE;
            }
            else if( read_data == 0x01 )
            {
              cam_ctrl->af_result = TRUE;
            }

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_NONE );

            rc = L_cam006_kernel_flash_off();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_flash_off\n");
                goto v_l_pre_flash_af_err_end;
            }

            cam_wake_up();
        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_YUV_STILL_MAIN_FLASH_OFF_AF_OFF )
    {

        if( cam_ctrl->main_flash_on_flg != TRUE )
        {
            cam_ctrl->yuv_main_flash_on_v_cnt++;
            if( cam_ctrl->yuv_main_flash_on_v_cnt >= 2 )
            {

                if( cam_ctrl->flash_parm_info.main_flash_on_flg == TRUE )
                {
#ifdef FLASH_TIMING_DEBUG
                    gpio_direction_output( 154, 1 );
#endif

                    rc = L_cam006_kernel_main_flash();
                    if( rc < 0 )
                    {
                        ERR_LOG("Timer TaskLet L_cam006_kernel_main_flash() Err\n");
                        goto v_l_still_err_end;
                    }
                }

                cam_ctrl->main_flash_on_flg = TRUE;

                cam_ctrl->yuv_main_flash_on_v_cnt = 0;
            }
            else
            {
                return;
            }
        }

        cam_ctrl->main_flash_off_v_cnt++;

        if(( cam_ctrl->main_flash_off_v_cnt >= 5 ) && ( cam_ctrl->main_flash_on_flg == TRUE ))
        {

            cam_ctrl->main_flash_off_v_cnt = 0;

            cam_ctrl->main_flash_on_flg = FALSE;

            cam_ctrl->pre_flash_on_flg = FALSE;

            L_cam006_kernel_set_vfall_free_irq();

#ifdef FLASH_TIMING_DEBUG
            gpio_direction_output( 154, 0 );
#endif

            rc = L_cam006_kernel_get_shutter_speed();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_get_shutter_speed Err\n");
            }

            rc = L_cam006_kernel_flash_off();
            if( rc < 0 )
            {
                ERR_LOG("Vsync Fall YUV Stil AF OFF Main Falsh Off Err\n");
                goto v_l_still_err_end;
            }

            rc = L_cam006_kernel_still_flash_end();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_still_flash_end() Err\n");
            }

        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_YUV_STILL_MAIN_FLASH_OFF_AF_ON )
    {

        if( cam_ctrl->main_flash_on_flg != TRUE )
        {
            cam_ctrl->yuv_main_flash_on_v_cnt++;
            if( cam_ctrl->yuv_main_flash_on_v_cnt >= 1 )
            {

                if( cam_ctrl->flash_parm_info.main_flash_on_flg == TRUE )
                {
#ifdef FLASH_TIMING_DEBUG
                    gpio_direction_output( 154, 1 );
#endif

                    rc = L_cam006_kernel_main_flash();
                    if( rc < 0 )
                    {
                      ERR_LOG("Timer TaskLet L_cam006_kernel_main_flash() Err\n");
                      goto v_l_still_err_end;
                    }
                }

                cam_ctrl->main_flash_on_flg = TRUE;

                cam_ctrl->yuv_main_flash_on_v_cnt = 0;
            }
            else
            {
                return;
            }
        }

        cam_ctrl->main_flash_off_v_cnt++;
        if(( cam_ctrl->main_flash_off_v_cnt >= 5 ) && ( cam_ctrl->main_flash_on_flg == TRUE ))
        {

            cam_ctrl->main_flash_off_v_cnt = 0;

            cam_ctrl->main_flash_on_flg = FALSE;

            cam_ctrl->pre_flash_on_flg = FALSE;

            L_cam006_kernel_set_vfall_free_irq();

#ifdef FLASH_TIMING_DEBUG
            gpio_direction_output( 154, 0 );
#endif

            rc = L_cam006_kernel_get_shutter_speed();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_get_shutter_speed Err\n");
            }

            rc = L_cam006_kernel_flash_off();
            if( rc < 0 )
            {
                ERR_LOG("Vsync Fall YUV Stil AF OFF Main Falsh Off Err\n");
            }

            rc = L_cam006_kernel_still_flash_end();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_still_flash_end() Err\n");
            }

        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_JPEG_STILL_NORMAL )
    {

        cam_ctrl->normal_still_v_cnt++;

        normal_v_check_cnt = 4;

        if( cam_ctrl->normal_still_v_cnt >= normal_v_check_cnt )
        {
            cam_ctrl->normal_still_v_cnt = 0;

            L_cam006_kernel_set_vfall_free_irq();

            rc = L_cam006_kernel_get_shutter_speed();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_get_shutter_speed Err\n");
            }

        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_JPEG_STILL_MAIN_FLASH_OFF_AF_OFF )
    {

        cam_ctrl->main_flash_off_v_cnt++;

        if(( cam_ctrl->main_flash_on_flg != TRUE ) && ( cam_ctrl->main_flash_off_v_cnt >= 2 ))
        {

            if( cam_ctrl->flash_parm_info.main_flash_on_flg == TRUE )
            {
#ifdef FLASH_TIMING_DEBUG
                gpio_direction_output( 154, 1 );
#endif

                rc = L_cam006_kernel_main_flash();
                if( rc < 0 )
                {
                    ERR_LOG("Timer TaskLet L_cam006_kernel_main_flash() Err\n");
                    goto v_l_still_err_end;
                }
            }

            cam_ctrl->main_flash_on_flg = TRUE;
        }

        if(( cam_ctrl->main_flash_on_flg == TRUE ) && ( cam_ctrl->main_flash_off_v_cnt >= 4 ))
        {
            cam_ctrl->main_flash_off_v_cnt = 0;
            cam_ctrl->main_flash_on_flg = FALSE;

            cam_ctrl->pre_flash_on_flg = FALSE;

            L_cam006_kernel_set_vfall_free_irq();

#ifdef FLASH_TIMING_DEBUG
            gpio_direction_output( 154, 0 );
#endif

            rc = L_cam006_kernel_get_shutter_speed();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_get_shutter_speed Err\n");
            }

            rc = L_cam006_kernel_flash_off();
            if( rc < 0 )
            {
                ERR_LOG("Flash Off Err\n");
                goto v_l_still_err_end;
            }

            rc = L_cam006_kernel_still_flash_end();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_still_flash_end() Err\n");
            }

            rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x06, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG("Capture Mode Change OK But 0x00FC Write Err\n");
                goto v_l_still_err_end;
            }
        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_JPEG_STILL_MAIN_FLASH_ON_AF_ON )
    {

        cam_ctrl->main_flash_off_v_cnt++;

        flash_on_v_check_cnt = 1;
        flash_off_v_check_cnt = 3;

        if(( cam_ctrl->main_flash_on_flg != TRUE ) && ( cam_ctrl->main_flash_off_v_cnt >= flash_on_v_check_cnt ))
        {

            if( cam_ctrl->flash_parm_info.main_flash_on_flg == TRUE )
            {
#ifdef FLASH_TIMING_DEBUG
                gpio_direction_output( 154, 1 );
#endif

                rc = L_cam006_kernel_main_flash();
                if( rc < 0 )
                {
                    ERR_LOG("Timer TaskLet L_cam006_kernel_main_flash() Err\n");
                    goto v_l_still_err_end;
                }
            }

            cam_ctrl->main_flash_on_flg = TRUE;
        }

        if(( cam_ctrl->main_flash_on_flg == TRUE ) && ( cam_ctrl->main_flash_off_v_cnt >= flash_off_v_check_cnt ))
        {

            cam_ctrl->main_flash_off_v_cnt = 0;

            cam_ctrl->main_flash_on_flg = FALSE;

            cam_ctrl->pre_flash_on_flg = FALSE;

            L_cam006_kernel_set_vfall_free_irq();

            rc = L_cam006_kernel_get_shutter_speed();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_get_shutter_speed Err\n");
            }

            rc = L_cam006_kernel_flash_off();
            if( rc < 0 )
            {
                ERR_LOG("Vsync Fall YUV Stil AF OFF Main Falsh Off Err\n");
                goto v_l_still_err_end;
            }

#ifdef FLASH_TIMING_DEBUG
            gpio_direction_output( 154, 0 );
#endif

            rc = L_cam006_kernel_still_flash_end();
            if( rc < 0 )
            {
                ERR_LOG("L_cam006_kernel_still_flash_end() Err\n");
            }
        }
    }

    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE )
    {

        rc = L_cam006_kernel_get_now_sensor_mode( &read_data );
        if( rc < 0 )
        {
            ERR_LOG("CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE Get Curret Mode Err\n");
            goto err_end;
        }

        if(( read_data & 0x03 ) == 0 )
        {

            L_cam006_kernel_set_vfall_free_irq();

            cam_wake_up();
        }
        else
        {

            rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x02, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE Err\n");
                goto err_end;
            }

            rc = cam_i2c_write_table( &cam_regs.still_to_preview_reg_settings[0],
                                           cam_regs.still_to_preview_reg_settings_size );
            if( rc < 0 )
            {
                ERR_LOG("CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE Err\n");
                goto err_end;
            }

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW );
        }
    }
    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW )
    {

        rc = L_cam006_kernel_get_now_sensor_mode( &read_data );
        if( rc < 0 )
        {
            ERR_LOG("CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW Get Now Sensor Mode Err\n");
            goto err_end;
        }

        if(( read_data & 0x03 ) == 0 )
        {

            rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x02, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW Err\n");
                goto err_end;
            }

            L_cam006_kernel_set_vfall_free_irq();

            cam_wake_up();
        }
    }
    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_DEFAULT_FOCUS )
    {

        rc = cam_i2c_read( cam_client->addr, 0x6D7A, &read_data, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS 0x6D7A Read Err\n");
            goto err_end;
        }

        if( read_data == 0 )
        {

            L_cam006_kernel_set_vfall_free_irq();

            cam_wake_up();
        }
        else
        {

            rc = cam_i2c_write( cam_client->addr, 0x4876, 0x0000, WORD_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS 0x4876 Write Err\n");
                goto err_end;
            }

            rc = cam_i2c_write( cam_client->addr, 0x002E, 0x02, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS 0x002E Write Err\n");
                goto err_end;
            }

            rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x02, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS Err\n");
                goto err_end;
            }

            rc = L_cam006_kernel_get_now_sensor_mode( &read_data );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS NowMoniterMode Read Err\n");
                goto err_end;
            }

            if(( read_data & 0x03 ) == 0 )
            {

                rc = cam_i2c_write( cam_client->addr, 0x0012, 0x01, ADWORD_DATBYTE_LEN );
                if( rc < 0 )
                {
                    ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS 0x0012 Write Err\n");
                    goto err_end;
                }
            }
            else
            {
                rc = cam_i2c_write( cam_client->addr, 0x0011, 0x00, ADWORD_DATBYTE_LEN );
                if( rc < 0 )
                {
                    ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS ModeSel1 Err\n");
                    goto err_end;
                }
            }

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK );
        }
    }
    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK )
    {

        rc = cam_i2c_read( cam_client->addr, 0x00F8, &read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0)
        {
            ERR_LOG("CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK Read Sensor Mode Check Err\n");
            goto err_end;
        }

        if( read_data & 0x02 )
        {

            rc = cam_i2c_write( cam_client->addr, 0x4852, 0x0000, WORD_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK Monitering reflesh Err\n");
                goto err_end;
            }

            rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x12, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK Err\n");
                goto err_end;
            }

            rc = cam_i2c_write( cam_client->addr, 0x4850, 0x01, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "CAMERA_V_FALL_DEFAULT_FOCUS_MODE_CHECK Manual Focus Err\n");
                goto err_end;
            }

            L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_DEFAULT_FOCUS_MOVE );
        }
    }
    else if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_DEFAULT_FOCUS_MOVE )
    {

        rc = cam_i2c_read( cam_client->addr, 0x00F8, &read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0)
        {
            ERR_LOG("CAMERA_V_FALL_DEFAULT_FOCUS_MOVE Read Focus Move Check Err\n");
            goto err_end;
        }

        if( read_data & 0x10 )
        {

            L_cam006_kernel_set_vfall_free_irq();

            if( cam_ctrl->af_proc_call_flg == TRUE )
            {
                cam_ctrl->af_proc_call_flg = FALSE;
            }

            cam_wake_up();
        }
    }

    return;

v_l_pre_flash_af_err_end:
  cam_ctrl->af_result = FALSE;

v_l_still_err_end:
  cam_ctrl->still_err_flg = FALSE;

err_end:
  L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_NONE );

  cam_wake_up();
  return;
}
DECLARE_TASKLET( cam_tasklet_vsync_low, cam_tasklet_proc_low, ( unsigned long )NULL );

static void cam_watch_wb_core(void)
{
    int rc = TRUE;

    uint16_t read_028A_NOW;
    int16_t read_0286_NOW;
    uint16_t read_440A_AEERR;
    uint16_t target_028A_NOW;

    rc = cam_i2c_read( cam_client->addr, 0x028A, &read_028A_NOW, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam_tasklet_proc_watch 0x028A Read Err\n");
        return;
    }

    rc = cam_i2c_read( cam_client->addr, 0x0286, &read_0286_NOW, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam_tasklet_proc_watch 0x0286 Read Err\n");
        return;
    }

    if(read_0286_NOW < 0)
    {
        read_0286_NOW *= -1;
    }

    if(cam_ctrl->power_on_type == CAMERA_POWER_ON_NORMAL)
    {
        switch(cam_ctrl->current_scene)
        {
            case CAM_SCENE_AUTO:
            case CAM_SCENE_PORTRAIT:
            case CAM_SCENE_SCENERY:
            case CAM_SCENE_ACTION:
            default:

                target_028A_NOW = 26811;
                break;
            case CAM_SCENE_NIGHT_PORTRAIT:
            case CAM_SCENE_NIGHT_SCENERY:

                target_028A_NOW = 25787;
                break;
        }
    }
    else
    {

        target_028A_NOW = 26245;
    }

    do
    {

        if(read_028A_NOW != target_028A_NOW ||
           read_0286_NOW < 400)
        {
            cam_ctrl->aw_watch_cnt = 0;
            break;
        }

        if(cam_ctrl->aw_watch_cnt >= 2)
        {
            rc = cam_i2c_read( cam_client->addr, 0x440A, &read_440A_AEERR, WORD_LEN );
            if( rc < 0 )
            {
                ERR_LOG("cam_tasklet_proc_watch Read INIT_AEERR Err\n");
                return;
            }

            if(read_440A_AEERR == 0x0400)
            {

                rc = cam_i2c_write( cam_client->addr, 0x440A, 0x1200, WORD_LEN );
                if( rc < 0 )
                {
                    ERR_LOG("cam_tasklet_proc_watch Write INIT_AEERR Err\n");
                    return;
                }
            }
        }
        else
        {
            cam_ctrl->aw_watch_cnt += 1;
            break;
        }
        return;
    }
    while(0);

    rc = cam_i2c_read( cam_client->addr, 0x440A, &read_440A_AEERR, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("cam_tasklet_proc_watch Read INIT_AEERR Err\n");
        return;
    }

    if(read_440A_AEERR == 0x1200)
    {

        rc = cam_i2c_write( cam_client->addr, 0x440A, 0x0400, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG("cam_tasklet_proc_watch Write INIT_AEERR Err\n");
            return;
        }
    }
}

static void cam_tasklet_proc_watch( unsigned long data )
{

    if(down_trylock(&cam_sem) != 0){

        return;
    }

    cam_watch_wb_core();

    up(&cam_sem);

}

DECLARE_TASKLET( cam_tasklet_vsync_watch, cam_tasklet_proc_watch, ( unsigned long )NULL );

static irqreturn_t vsync_high_interrupt(int irq, void *handle)
{
  

  if(cam_ctrl->v_rise_proc != CAMERA_V_RISE_NONE)
  {
    tasklet_schedule( &cam_tasklet_vsync_high );

  }
  else
  {
    tasklet_schedule( &cam_tasklet_vsync_watch );
  }

  return IRQ_HANDLED;
}

static irqreturn_t vsync_low_interrupt(int irq, void *handle)
{
  

  if(cam_ctrl->v_fall_proc != CAMERA_V_FALL_NONE)
  {
    tasklet_schedule( &cam_tasklet_vsync_low );

  }
  else
  {
    tasklet_schedule( &cam_tasklet_vsync_watch );
  }

  return IRQ_HANDLED;
}

static int32_t cam_i2c_txdata(unsigned short saddr,
    unsigned char *txdata, int length)
{
    struct i2c_msg msg[] = {
        {
            .addr = saddr,
            .flags = 0,
            .len = length,
            .buf = txdata,
        },
    };

#ifdef I2C_DEBUG
    unsigned short addr ;
    unsigned char *data ;

    data = txdata ;
    switch(length)
    {
    case 4 :
      addr = *data ;
      data++ ;
      addr = addr << 8  ;
      addr += *data ;
      data++ ;

      break ;
    case 2:
      addr = *data ;
      data++ ;

      break ;
    case 3:
      addr = *data ;
      data++ ;
      addr = addr << 8  ;
      addr += *data ;
      data++ ;

      break ;
    case 6:
      addr = *data ;
      data++ ;
      addr = addr << 8  ;
      addr += *data ;
      data++ ;

      break ;
    }
#endif

    if (i2c_transfer(cam_client->adapter, msg, 1) < 0) {
        ERR_LOG("cam_i2c_txdata faild\n");
        return -EIO;
    }

    return 0;
}

static int32_t cam_i2c_write(unsigned short saddr,
    unsigned short waddr, unsigned long wdata, enum cam_width_t width)
{
    int32_t rc = -EFAULT;
    unsigned char buf[6];

    memset(buf, 0, sizeof(buf));

    switch (width) {
    case WORD_LEN: {
      buf[0] = (unsigned char)((waddr & 0xFF00)>>8);
      buf[1] = (unsigned char)(waddr & 0x00FF);
      buf[2] = (unsigned char)(wdata & 0x000000FF);
      buf[3] = (unsigned char)((wdata & 0x0000FF00)>>8);

      rc = cam_i2c_txdata(saddr, buf, 4);
    }
      break;

    case BYTE_LEN: {
      buf[0] = (unsigned char)(waddr & 0x000000FF);
      buf[1] = (unsigned char)(wdata & 0x000000FF);

      rc = cam_i2c_txdata(saddr, buf, 2);
    }
      break;

    case ADWORD_DATBYTE_LEN: {
      buf[0] = (unsigned char)((waddr & 0xFF00)>>8);
      buf[1] = (unsigned char)(waddr & 0x00FF);
      buf[2] = (unsigned char)(wdata & 0x000000FF);

      rc = cam_i2c_txdata(saddr, buf, 3);
    }
      break;

    case ADWORD_DATLONG_LEN: {
      buf[0] = (unsigned char)((waddr & 0xFF00)>>8);
      buf[1] = (unsigned char)(waddr & 0x00FF);
      buf[2] = (unsigned char)((wdata & 0x000000FF));
      buf[3] = (unsigned char)((wdata & 0x0000FF00)>>8);
      buf[4] = (unsigned char)((wdata & 0x00FF0000)>>16);
      buf[5] = (unsigned char)((wdata & 0xFF000000)>>24);

      rc = cam_i2c_txdata(saddr, buf, 6);
    }
      break;

    default: {

      rc = -1;
    }
        break;

    }

    if (rc < 0)
        ERR_LOG("f A[0x%04x]D[0x%04x]!\n", waddr, (unsigned int)wdata);

    return rc;
}

static int32_t cam_i2c_write_table(
    struct cam_i2c_reg_conf const *reg_conf_tbl,
    int num_of_items_in_table)
{
    int i;
    int32_t rc = -EFAULT;

    for (i = 0; i < num_of_items_in_table; i++) {
        rc = cam_i2c_write(cam_client->addr,
            reg_conf_tbl->waddr, reg_conf_tbl->wdata,
            reg_conf_tbl->width);
      if (rc < 0) {

            break;
      }
      if (reg_conf_tbl->mdelay_time != 0) {

            mdelay(reg_conf_tbl->mdelay_time);
      }
        reg_conf_tbl++;
    }

    return rc;
}

#ifdef CAMSENSOR_DEBUG
static int32_t cam_i2c_write_DEBUG_table(
    struct cam_i2c_reg_debug_conf const *reg_conf_tbl,
    int num_of_items_in_table)
{
    int i;
    int32_t rc = -EFAULT;
  unsigned short page_addr = 0x0000;
  unsigned short reg_addr = 0x0000;
  unsigned short rdata = 0x0000;

    for (i = 0; i < num_of_items_in_table; i++) {
      page_addr = (reg_conf_tbl->waddr >> 16) & 0x0000FFFF;
      reg_addr  = reg_conf_tbl->waddr & 0x0000FFFF;

      if(page_addr != 0xFCFC) {
        rc = cam_i2c_write(cam_client->addr, 0x0028, page_addr,reg_conf_tbl->width);
        rc = cam_i2c_write(cam_client->addr, 0x002A, reg_addr,reg_conf_tbl->width);
        rc = cam_i2c_write(cam_client->addr, 0x0F12, reg_conf_tbl->wdata,reg_conf_tbl->width);

        rdata = 0xFFFF;
      }
      else
      {
        rdata = 0xFFFF;
        rc = cam_i2c_write(cam_client->addr, page_addr, reg_addr,reg_conf_tbl->width);
        MSG_FATAL("r A[0x%04x]D[0x%04x]P[----]\n", page_addr,rdata,0x0000);
      }
      if (rc < 0) {

            break;
      }
      if (reg_conf_tbl->mdelay_time != 0) {

            mdelay(reg_conf_tbl->mdelay_time);
      }
        reg_conf_tbl++;
    }

    return rc;
}

static int32_t cam_i2c_read_DEBUG_table(
    struct cam_i2c_reg_debug_conf const *reg_conf_tbl,
    int num_of_items_in_table)
{
    int i;
    int32_t rc = -EFAULT;
  unsigned short page_addr = 0x0000;
  unsigned short reg_addr = 0x0000;
  unsigned short rdata = 0x0000;

    for (i = 0; i < num_of_items_in_table; i++) {
      page_addr = (reg_conf_tbl->waddr >> 16) & 0x0000FFFF;
      reg_addr  = reg_conf_tbl->waddr & 0x0000FFFF;

      if(page_addr != 0xFCFC) {
        rdata = 0xFFFF;
        rc = cam_i2c_write(cam_client->addr, 0x002C, page_addr,reg_conf_tbl->width);
        rc = cam_i2c_write(cam_client->addr, 0x002E, reg_addr,reg_conf_tbl->width);
        rc = cam_i2c_read(cam_client->addr, 0x0F12, &rdata,reg_conf_tbl->width);
        MSG_FATAL("r A[0x%04x]D[0x%04x]P[0x%04x]\n", reg_addr,rdata,page_addr);
      }
      else
      {
        rdata = 0xFFFF;
        rc = cam_i2c_read(cam_client->addr, page_addr, &rdata,reg_conf_tbl->width);
        MSG_FATAL("r A[0x%04x]D[0x%04x]P[----]\n", page_addr,rdata,0x0000);
      }
      if (rc < 0) {

            break;
      }
      if (reg_conf_tbl->mdelay_time != 0) {

            mdelay(reg_conf_tbl->mdelay_time);
      }
        reg_conf_tbl++;
    }

    return rc;
}
#endif

static int cam_i2c_rxdata(unsigned short saddr,
    unsigned char *rxdata, int length)
{

#ifdef I2C_DEBUG
    unsigned short addr ;
    unsigned char *data ;
#endif

    struct i2c_msg msgs[] = {
    {
        .addr   = saddr,
        .flags = 0,
        .len   = 2,
        .buf   = rxdata,
    },
    {
        .addr   = saddr,
        .flags = I2C_M_RD,
        .len   = length,
        .buf   = rxdata,
    },
    };

#ifdef I2C_DEBUG
    data = rxdata ;
    addr = *data ;
    data++ ;
    addr = addr << 8  ;
    addr += *data ;
#endif

    if (i2c_transfer(cam_client->adapter, msgs, 2) < 0) {
        ERR_LOG("cam_i2c_rxdata failed!\n");
        return -EIO;
    }

#ifdef I2C_DEBUG
    data = rxdata ;
    switch(length)
    {
    case 1 :

      break ;
    case 2 :

      break ;
    case 4 :

      break ;
    }
#endif

    return 0;
}

static int32_t cam_i2c_read(unsigned short   saddr,
    unsigned short raddr, unsigned short *rdata, enum cam_width_t width)
{
    int32_t rc = 0;
    unsigned char buf[4];
    unsigned long read_data = 0;

    if (!rdata)
        return -EIO;

    memset(buf, 0, sizeof(buf));

    *rdata = 0;

    switch (width) {
    case WORD_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
        buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 2);
        if (!(rc < 0))

            *rdata = buf[1] << 8 | buf[0];
    }
      break;

    case BYTE_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
        buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 1);
        if (!(rc < 0))
            *rdata = buf[0];
    }
      break;

    case ADWORD_DATBYTE_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
        buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 1);
        if (!(rc < 0))
            *rdata = buf[0];
    }
      break;

    case ADWORD_DATLONG_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
        buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 4);
        if (!(rc < 0))
        {

            read_data = (( buf[3] << 24) | ( buf[2] << 16 ) | ( buf[1] << 8 ) | ( buf[0] )) ;
        }
    }
      break;

    default: {

      rc = -1;
        break;
    }
    }

    if (rc < 0)
        ERR_LOG("f A[0x%04x]D[0x%04x]!\n", raddr, *rdata);

    return rc;
}

static int32_t cam_i2c_read_long(unsigned short   saddr,
    unsigned short raddr, unsigned long *rdata, enum cam_width_t width)
{
    int32_t rc = 0;
    unsigned char buf[4];

    if (!rdata)
        return -EIO;

    memset(buf, 0, sizeof(buf));

    *rdata = 0;

    switch (width) {
    case WORD_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
      buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 2);
        if (!(rc < 0))

            *rdata = buf[1] << 8 | buf[0];
    }
      break;

    case BYTE_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
      buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 1);
        if (!(rc < 0))
            *rdata = buf[0];
    }
      break;

    case ADWORD_DATBYTE_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
      buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 1);
        if (!(rc < 0))
            *rdata = buf[0];
    }
      break;

    case ADWORD_DATLONG_LEN: {
        buf[0] = (raddr & 0xFF00)>>8;
      buf[1] = (raddr & 0x00FF);

        rc = cam_i2c_rxdata(saddr, buf, 4);
        if (!(rc < 0))
        {

            *rdata = (( buf[3] << 24) | ( buf[2] << 16 ) | ( buf[1] << 8 ) | ( buf[0] )) ;
        }
    }
      break;

    default: {

      rc = -1;
        break;
    }
    }

    if (rc < 0)
        ERR_LOG("f A[0x%04x]D[0x%04x]!\n", raddr, (unsigned int)*rdata);

    return rc;
}

#ifndef CONFIG_TARGET_PROTOTYPE_WS1

static int32_t cam_i2c_flash_write(unsigned short saddr,
    unsigned char data1, unsigned char data2, unsigned char data3 )
{
    int32_t rc = TRUE;
    unsigned char buf[3];

    memset(buf, 0, sizeof(buf));

    buf[0] = data1;
    buf[1] = data2;
    buf[2] = data3;

    rc = cam_i2c_txdata(saddr, buf, 3);

    if (rc < 0)
      printk("Err data1:0x%2x data2:0x%2x data3:0x%2x \n", data1, data2, data3 );
    else
      printk("Success data1:0x%2x data2:0x%2x data3:0x%2x \n", data1, data2, data3 );

    return rc;
}

#endif

static int cam_read_hw_ver( void )
{
  int rc = 0;

  cam_ctrl->hw_ver = 0x00000000;
  return rc;
}

static int cam_set_reg_table( void )
{
  int rc = 0;

  cam_ctrl->cam_regs = &cam_regs;

  return rc;
}

static long cam_reg_init(void)
{
    return 0;
}

static int cam006_kernel_search_init_reg_offset( struct cam_i2c_reg_conf *tbl, unsigned short waddr, uint16_t max_size )
{
  int index ;
  for(index = 0 ; index < max_size; index++)
  {
    if(tbl->waddr == waddr)
    {
      return index ;
    }
    tbl++ ;
  }
  return -1 ;
}

static int cam_sensor_init_probe(const struct msm_camera_sensor_info *data)
{
    int rc = 0;

  cam_preview_sxga_offset.offset_2541 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2541,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2367 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2367,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2368 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2368,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2370 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2370,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2369 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2369,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2371 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2371,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2406 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2406,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2521 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2521,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2407 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2407,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2373 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2373,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2374 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2374,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2375 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2375,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2376 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2376,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2377 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2377,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2378 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2378,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;
  cam_preview_sxga_offset.offset_2362 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_sxga_init_reg_settings_4,
                  REG_OFFSET_2362,
                  cam_regs.preview_sxga_init_reg_settings_4_size ) ;

  cam_preview_720p_offset.offset_2541 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2541,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2367 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2367,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2368 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2368,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2370 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2370,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2369 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2369,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2371 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2371,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2406 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2406,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2521 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2521,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2407 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2407,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2373 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2373,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2374 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2374,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2375 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2375,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2376 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2376,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2377 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2377,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2378 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2378,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;
  cam_preview_720p_offset.offset_2362 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_720p_init_reg_settings_4,
                  REG_OFFSET_2362,
                  cam_regs.preview_720p_init_reg_settings_4_size ) ;

  cam_preview_wvga_offset.offset_2541 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2541,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2367 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2367,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2368 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2368,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2370 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2370,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2369 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2369,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2371 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2371,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2406 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2406,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2521 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2521,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2407 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2407,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2373 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2373,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2374 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2374,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2375 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2375,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2376 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2376,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2377 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2377,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2378 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2378,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;
  cam_preview_wvga_offset.offset_2362 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_wvga_init_reg_settings_4,
                  REG_OFFSET_2362,
                  cam_regs.preview_wvga_init_reg_settings_4_size ) ;

  cam_preview_vga_offset.offset_2541 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2541,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2367 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2367,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2368 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2368,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2370 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2370,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2369 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2369,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2371 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2371,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2406 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2406,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2521 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2521,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2407 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2407,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2373 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2373,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2374 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2374,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2375 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2375,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2376 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2376,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2377 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2377,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2378 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2378,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;
  cam_preview_vga_offset.offset_2362 =
          cam006_kernel_search_init_reg_offset(
                  cam_regs.preview_vga_init_reg_settings_4,
                  REG_OFFSET_2362,
                  cam_regs.preview_vga_init_reg_settings_4_size ) ;

    return rc;
}

int cam_sensor_init(const struct msm_camera_sensor_info *data)
{
    int rc = 0;

    cam_ctrl = kzalloc(sizeof(struct cam_ctrl_t), GFP_KERNEL);
    if (!cam_ctrl) {
        
        rc = -ENOMEM;
        goto kzalloc_fail;
    }

  if (data)
    cam_ctrl->sensordata = data;
  else {
    
    rc = -1;
    goto init_fail;
  }
  cam_ctrl->cam_regs       = NULL;
  cam_ctrl->wait_flag          = 0;
  cam_ctrl->sensor_state       = 0x00000000;
  cam_ctrl->preview_mode       = PREV_STILL_MODE;
  cam_ctrl->still_mode         = STILL_JPEG_MODE;
  cam_ctrl->flash_mode         = FLASH_OFF_MODE;
  cam_ctrl->af_mode            = AF_OFF_MODE;
  cam_ctrl->vref_high_int_func = 0x00000000;
  cam_ctrl->vref_low_int_func  = 0x00000000;
  cam_ctrl->vsync_count        = 0;
  cam_ctrl->af_cmd.cmd         = MAX_AF_CMD;
  cam_ctrl->af_cmd.param       = 0x0000;
  cam_ctrl->gfps.prevfps       = 0;
  cam_ctrl->gfps.pictfps       = 0;
  cam_ctrl->hw_ver             = 0x00000000;

  cam_ctrl->current_picture_width      = DEFAULT_STILL_5M_WIDTH;
  cam_ctrl->current_picture_height     = DEFAULT_STILL_5M_HEIGHT;
  cam_ctrl->current_display_width      = DEFAULT_PREVIEW_SIZE_WIDTH;
  cam_ctrl->current_display_height     = DEFAULT_PREVIEW_SIZE_HEIGHT;
  cam_ctrl->past_picture_width = DEFAULT_STILL_5M_WIDTH;
  cam_ctrl->past_picture_height = DEFAULT_STILL_5M_HEIGHT;
  cam_ctrl->past_display_width = DEFAULT_PREVIEW_SIZE_WIDTH;
  cam_ctrl->past_display_height = DEFAULT_PREVIEW_SIZE_HEIGHT;
  cam_ctrl->camera_state = CAMERA_INIT_STATE;
  cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;
  cam_ctrl->current_effect = CAM_EFFECT_NONE;
  cam_ctrl->current_wb = CAM_WB_AUTO;
  cam_ctrl->current_iso = CAM_ISO_AUTO;
  cam_ctrl->current_ae_mode = CAM_AEC_AVERAGE;

  cam_ctrl->current_wb_update = TRUE;
  cam_ctrl->current_iso_update = TRUE;
  cam_ctrl->current_ae_mode_update = TRUE;

  cam_ctrl->current_afwin_info.width      = AF_WINDOW_WIDTH_DEF;
  cam_ctrl->current_afwin_info.height     = AF_WINDOW_HEIGHT_DEF;
  cam_ctrl->current_afwin_info.pos_x      = AF_WINDOW_POSX_DEF;
  cam_ctrl->current_afwin_info.pos_y      = AF_WINDOW_POSY_DEF;
  cam_ctrl->current_aewin_info.width      = AE_WINDOW_WIDTH_DEF;
  cam_ctrl->current_aewin_info.height     = AE_WINDOW_HEIGHT_DEF;
  cam_ctrl->current_awbwin_info.width     = AWB_WINDOW_WIDTH_DEF;
  cam_ctrl->current_awbwin_info.height    = AWB_WINDOW_HEIGHT_DEF;

  cam_ctrl->current_fps = FPS_15;
  cam_ctrl->past_fps = FPS_15;

  cam_ctrl->current_lens_shading = TRUE;

  cam_ctrl->current_scene = CAM_SCENE_AUTO;

  cam_ctrl->scene_err_flg = FALSE;

  cam_ctrl->current_scene_update = TRUE;

  cam_ctrl->still_err_flg = FALSE;
  memset( &cam_ctrl->flash_parm_info, 0x00, sizeof( camera_flash_parm_type ));
  cam_ctrl->af_proc_call_flg = FALSE;

  cam_ctrl->pre_flash_on_flg = FALSE;

  cam_ctrl->main_flash_on_flg = FALSE;
  cam_ctrl->main_flash_off_v_cnt = 0;
  cam_ctrl->yuv_main_flash_on_v_cnt = 0;

 memset( &cam_ctrl->cam006_af_otp_data, 0x00, sizeof( camera_set_af_otp_data_type ));

  cam_ctrl->internal_power_offon_retry_flg = FALSE;

  cam_ctrl->normal_still_v_cnt = 0;

#ifdef TIME_CHECK
  do_gettimeofday(&cam_ctrl->vsync_high_tv);
  do_gettimeofday(&cam_ctrl->prev_vsync_high_tv);
  do_gettimeofday(&cam_ctrl->vsync_low_tv);
  do_gettimeofday(&cam_ctrl->prev_vsync_low_tv);
  do_gettimeofday(&cam_ctrl->seq_tv);
  do_gettimeofday(&cam_ctrl->prev_seq_tv);
#endif

  free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

  cam_ctrl->aw_watch_cnt = 0;
  L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_NONE );
  L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

#ifdef _FLASH_ENABLE_
  if(!cam628_client) {
    cam628_init_probe();
    cam628_set_data(data);
  }
#endif

  cam_ctrl->flash_on_flg = 0;

  rc = cam_set_reg_table();
  if (rc < 0) {
    
    goto init_fail;
  }

  rc = L_cam006_kernel_sensor_power_on( CAMERA_POWER_ON_NORMAL );
  if( rc < 0 )
  {
    
    rc = -1;
    goto init_fail;
  }
  msm_camio_camif_pad_reg_reset();

  rc = cam_read_hw_ver();
  if (rc < 0) {
    
    goto init_fail;
  }

  rc = cam_reg_init();
    if (rc < 0) {
        
        goto init_fail;
    }

  rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_high_interrupt, IRQF_TRIGGER_RISING, "cam_vsync_h", 0);
  if (rc < 0) {
    
    goto init_fail;
  }

  cam_chg_state(SENSOR_STATE_MODE, SENSOR_ON);

  cam_ctrl->camera_state = CAMERA_POWERON_STATE;

    return rc;

kzalloc_fail:
    return rc;
init_fail:
    kfree(cam_ctrl);
    return rc;
}

static int cam_vsync_wait(unsigned int func_flag, unsigned int v_type)
{

  long rc = 0;

  cam_ctrl->vsync_count = 0;

  switch(func_flag)
  {
    case MODE_PREV_VREF_INT:
      if(v_type == VSYNC_HIGH)
        cam_ctrl->vref_high_int_func |= MODE_PREV_VREF_INT;
      else
        cam_ctrl->vref_low_int_func |= MODE_PREV_VREF_INT;
      break;
    case MODE_STILL_VREF_INT:
      if(v_type == VSYNC_HIGH)
        cam_ctrl->vref_high_int_func |= MODE_STILL_VREF_INT;
      else
        cam_ctrl->vref_low_int_func |= MODE_STILL_VREF_INT;
      break;
    case AF_VREF_INT:
      if(v_type == VSYNC_HIGH)
        cam_ctrl->vref_high_int_func |= AF_VREF_INT;
      else
        cam_ctrl->vref_low_int_func |= AF_VREF_INT;
      break;
    case PRE_FLASH_ON_VREF_INT:
      if(v_type == VSYNC_HIGH)
        cam_ctrl->vref_high_int_func |= PRE_FLASH_ON_VREF_INT;
      else
        cam_ctrl->vref_low_int_func |= PRE_FLASH_ON_VREF_INT;
      break;
    case FLASH_OFF_VREF_INT:
      if(v_type == VSYNC_HIGH)
        cam_ctrl->vref_high_int_func |= FLASH_OFF_VREF_INT;
      else
        cam_ctrl->vref_low_int_func |= FLASH_OFF_VREF_INT;
      break;
    default:
      return 0;
  }

#ifdef TIME_CHECK
  do_gettimeofday(&cam_ctrl->prev_seq_tv);
#endif

  rc = cam_wait_event();

#ifdef TIME_CHECK
  get_vsync_time(&cam_ctrl->seq_tv, &cam_ctrl->prev_seq_tv);
#endif

  switch(func_flag)
  {
    case MODE_PREV_VREF_INT:
      
      break;
    case MODE_STILL_VREF_INT:
      
      break;
    case AF_VREF_INT:
      break;
    case PRE_FLASH_ON_VREF_INT:
      break;
    case FLASH_OFF_VREF_INT:
      break;
    default:
      return 0;
  }

  return rc;
}

static int L_cam006_kernel_sensor_power_on( camera_power_on_reg_type power_on_type )
{
    int rc = TRUE;

    gpio_direction_output( 156, 1 );
    mdelay(2);

    HS_CAM_VDD_IO_ON();
    mdelay(1);

    HS_CAM_VDD_A_ON();
    mdelay(1);

    HS_CAM_VDD_AF_ON();
    mdelay(1);

    msm_camio_clk_rate_set(24000000);
    gpio_tlmm_config(GPIO_CFG(15, 1, GPIO_CFG_OUTPUT, GPIO_CFG_NO_PULL, GPIO_CFG_4MA), GPIO_CFG_ENABLE);
    mdelay(5);

    gpio_direction_output( 3, 1 );
    mdelay(7);

    rc = cam006_check_internal_status( 0x01, 0x01 );
    if( rc < 0 )
    {
        ERR_LOG("cam006_check_internal_status PowerDown - PreSleep Err\n");
        return ERR_CODE;
    }

    cam_ctrl->power_on_type = power_on_type;

    rc = cam006_initial_setting_1( power_on_type );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_1 Err\n");
        return ERR_CODE;
    }

    mdelay(1);

    rc = cam006_check_internal_status( 0x01, 0x01 );
    if( rc < 0 )
    {
        ERR_LOG("cam006_check_internal_status PreSleep - Sleep Err\n");
        return ERR_CODE;
    }

    rc = cam006_o_data_read();
    if( rc < 0 )
    {
        ERR_LOG("cam006_o_data_read Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_2( power_on_type );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_2 Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_3();
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_3 Err\n");
        return ERR_CODE;
    }

#ifdef REG_C20P
#else
    rc = cam006_initial_write_rom_correction();
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_write_rom_correction Err\n");
        return ERR_CODE;
    }
#endif

    gpio_direction_output( 0, 1 );

    mdelay(18);

    rc = cam006_check_internal_status( 0x01, 0x03);
    if( rc < 0 )
    {
        ERR_LOG("cam006_check_internal_status Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_4( power_on_type );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_4 Err\n");
        return ERR_CODE;
    }

    rc = cam006_AFdriver_data_write();
    if( rc < 0 )
    {
        ERR_LOG("cam006_AFdriver_data_write Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_5( power_on_type );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_5 Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int L_cam006_kernel_sensor_power_off( void )
{

    gpio_direction_output( 0, 0 );
    mdelay(5);

    gpio_direction_output( 3, 0 );
    mdelay(1);

    gpio_tlmm_config(GPIO_CFG(15, 0, GPIO_CFG_OUTPUT, GPIO_CFG_PULL_DOWN, GPIO_CFG_4MA), GPIO_CFG_ENABLE);
    gpio_direction_output( 15, 0 );
    mdelay(1);

    HS_CAM_VDD_AF_OFF();
    mdelay(15);

    HS_CAM_VDD_A_OFF();
    mdelay(5);

    HS_CAM_VDD_IO_OFF();
    mdelay(3);

    tps65023_dcdc3_i2c_write( 0x01 );

    gpio_direction_output( 156, 0 );
    mdelay(50);

    return TRUE;
}

static int L_cam006_kernel_sensor_dlc_power_on( void )
{
    int rc = TRUE;

    gpio_direction_output( 156, 1 );
    mdelay(2);

    HS_CAM_VDD_IO_ON();
    mdelay(1);

    HS_CAM_VDD_A_ON();
    mdelay(1);

    HS_CAM_VDD_AF_ON();
    mdelay(1);

    msm_camio_clk_rate_set(24000000);
    gpio_tlmm_config(GPIO_CFG(15, 1, GPIO_CFG_OUTPUT, GPIO_CFG_NO_PULL, GPIO_CFG_4MA), GPIO_CFG_ENABLE);
    mdelay(5);

    gpio_direction_output( 3, 1 );
    mdelay(10);

    rc = cam006_check_internal_status( 0x01, 0x01 );
    if( rc < 0 )
    {
        ERR_LOG("cam006_check_internal_status PowerDown - PreSleep Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_1( CAMERA_POWER_ON_TEST_DL );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_1 Err\n");
        return ERR_CODE;
    }

    mdelay(1);

    rc = cam006_check_internal_status( 0x01, 0x01 );
    if( rc < 0 )
    {
        ERR_LOG("cam006_check_internal_status PreSleep - Sleep Err\n");
        return ERR_CODE;
    }

    rc = cam006_o_data_read();
    if( rc < 0 )
    {
        ERR_LOG("cam006_o_data_read Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_2( CAMERA_POWER_ON_TEST_DL );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_2 Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_3();
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_3 Err\n");
        return ERR_CODE;
    }

#ifdef REG_C20P
#else
    rc = cam006_initial_write_rom_correction();
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_write_rom_correction Err\n");
        return ERR_CODE;
    }
#endif

    rc = cam006_initial_setting_4( CAMERA_POWER_ON_TEST_DL );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_4 Err\n");
        return ERR_CODE;
    }

    rc = cam006_initial_setting_5( CAMERA_POWER_ON_TEST_DL );
    if( rc < 0 )
    {
        ERR_LOG("cam006_initial_setting_5 Err\n");
        return ERR_CODE;
    }

    gpio_direction_output( 0, 1 );

    mdelay(30);

    rc = cam006_check_internal_status( 0x01, 0x03);
    if( rc < 0 )
    {
        ERR_LOG("cam006_check_internal_status Err\n");
        return ERR_CODE;
    }

    rc = cam006_AFdriver_data_write();
    if( rc < 0 )
    {
        ERR_LOG("cam006_AFdriver_data_write Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int L_cam006_start_preview( void )
{
  int rc = 0;
  int power_off_on_preview_mode_chg_flg = FALSE;
  int power_off_on_720p_preview_mode_chg_flg = FALSE;
  int power_off_on_wvga_preview_mode_chg_flg = FALSE;
  camera_power_on_reg_type power_on_type;

  if( cam_ctrl->past_fps <= FPS_15 )
  {

    if( cam_ctrl->current_fps >= FPS_24 )
    {
        power_off_on_preview_mode_chg_flg = TRUE;
        if(( cam_ctrl->current_display_width == VIDEO_720P_WIDTH ) &&
           ( cam_ctrl->current_display_height == VIDEO_720P_HEIGHT ))
        {

            power_on_type = CAMERA_POWER_ON_MOVIE_720P;
        }
        else if(( cam_ctrl->current_display_width == VIDEO_WVGA_WIDTH ) &&
                ( cam_ctrl->current_display_height == VIDEO_WVGA_HEIGHT ))
        {

            power_on_type = CAMERA_POWER_ON_MOVIE_WVGA;
        }
        else if(( cam_ctrl->current_display_width <= VIDEO_VGA_WIDTH ) &&
                ( cam_ctrl->current_display_height <= VIDEO_VGA_HEIGHT ))
        {

            power_on_type = CAMERA_POWER_ON_MOVIE_VGA;
        }
        else
        {

            power_on_type = CAMERA_POWER_ON_NORMAL;
        }
    }
  }

  else if( cam_ctrl->past_fps >= FPS_24 )
  {

    if( cam_ctrl->current_fps <= FPS_15 )
    {
      power_off_on_preview_mode_chg_flg = TRUE;
      power_on_type = CAMERA_POWER_ON_NORMAL;
    }

    else if( cam_ctrl->current_fps >= FPS_24 )
    {

      if((( cam_ctrl->past_display_width > VIDEO_VGA_WIDTH )    &&
         ( cam_ctrl->past_display_height >= VIDEO_VGA_HEIGHT ))  &&
         (( cam_ctrl->current_display_width <= VIDEO_VGA_WIDTH ) &&
          ( cam_ctrl->current_display_height <= VIDEO_VGA_HEIGHT )))
      {

         power_off_on_preview_mode_chg_flg = TRUE;
         power_on_type = CAMERA_POWER_ON_MOVIE_VGA;

      }

      else if((( cam_ctrl->past_display_width <= VIDEO_VGA_WIDTH )    &&
              ( cam_ctrl->past_display_height <= VIDEO_VGA_HEIGHT ))  &&
              (( cam_ctrl->current_display_width > VIDEO_VGA_WIDTH ) &&
               ( cam_ctrl->current_display_height >= VIDEO_VGA_HEIGHT )))
      {
        power_off_on_preview_mode_chg_flg = TRUE;

        if(( cam_ctrl->current_display_width == VIDEO_720P_WIDTH ) &&
           ( cam_ctrl->current_display_height == VIDEO_720P_HEIGHT ))
        {

            power_on_type = CAMERA_POWER_ON_MOVIE_720P;
        }
        else if(( cam_ctrl->current_display_width == VIDEO_WVGA_WIDTH ) &&
                ( cam_ctrl->current_display_height == VIDEO_WVGA_HEIGHT ))
        {

            power_on_type = CAMERA_POWER_ON_MOVIE_WVGA;
        }
      }
      else
      {

        if((( cam_ctrl->past_display_width == VIDEO_720P_WIDTH )    &&
            ( cam_ctrl->past_display_height == VIDEO_720P_HEIGHT )) &&
           (( cam_ctrl->current_display_width == VIDEO_WVGA_WIDTH ) &&
            ( cam_ctrl->current_display_height == VIDEO_WVGA_HEIGHT )))
        {

            power_off_on_preview_mode_chg_flg = FALSE;
            power_off_on_wvga_preview_mode_chg_flg = TRUE;
        }

        else if((( cam_ctrl->past_display_width == VIDEO_WVGA_WIDTH )    &&
                 ( cam_ctrl->past_display_height == VIDEO_WVGA_HEIGHT )) &&
                (( cam_ctrl->current_display_width == VIDEO_720P_WIDTH ) &&
                 ( cam_ctrl->current_display_height == VIDEO_720P_HEIGHT )))
        {

            power_off_on_preview_mode_chg_flg = FALSE;
            power_off_on_720p_preview_mode_chg_flg = TRUE;
        }
      }
    }
  }

  if( power_off_on_preview_mode_chg_flg == TRUE )
  {

    cam_ctrl->internal_power_offon_retry_flg = TRUE;

    L_cam006_kernel_sensor_power_off();

    rc = L_cam006_kernel_sensor_power_on( power_on_type );
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_start_preview L_cam006_kernel_sensor_power_on Err\n");
        cam_ctrl->internal_power_offon_retry_flg = FALSE;
        return ERR_CODE;
    }

    rc = L_cam006_kernel_still_movie_chg_retry_set_image_control(power_on_type);
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_start_preview Retry Set ImageControl Err\n");
        cam_ctrl->internal_power_offon_retry_flg = FALSE;
        return ERR_CODE;
    }

    cam_ctrl->internal_power_offon_retry_flg = FALSE;

  }
  else
  {

    if( power_off_on_720p_preview_mode_chg_flg == TRUE )
    {
        rc = cam_i2c_write_table( &cam_regs.preview_720p_reg_settings[0],
                                      cam_regs.preview_720p_reg_settings_size );
        if( rc < 0 )
        {
            ERR_LOG("L_cam006_start_preview 720p cam_i2c_write_table Err\n");
            return ERR_CODE;
        }

        rc = cam006_check_internal_status( 0x02, 0x02 );
        if( rc < 0 )
        {
            ERR_LOG("L_cam006_start_preview 720p cam006_check_internal_status Err\n");
            return ERR_CODE;
        }
    }
    else if( power_off_on_wvga_preview_mode_chg_flg == TRUE )
    {
        rc = cam_i2c_write_table( &cam_regs.preview_wvga_reg_settings[0],
                                      cam_regs.preview_wvga_reg_settings_size );
        if( rc < 0 )
        {
            ERR_LOG("L_cam006_start_preview WVGA cam_i2c_write_table Err\n");
            return ERR_CODE;
        }

        rc = cam006_check_internal_status( 0x02, 0x02 );
        if( rc < 0 )
        {
            ERR_LOG("L_cam006_start_preview WVGA cam006_check_internal_status Err\n");
            return ERR_CODE;
        }
    }
  }

  return rc;
}

static int L_cam006_kernel_move_focus_infinity( void )
{
    int rc = TRUE;

    uint16_t read_data;

    rc = cam_i2c_read( cam_client->addr, 0x6D7A, &read_data, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x6D7A Read Err\n");
        return ERR_CODE;
    }

    if( read_data != 0 )
    {

        rc = cam_i2c_write( cam_client->addr, 0x4876, 0x0000, WORD_LEN );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x4876 Write Err\n");
          return ERR_CODE;
        }

        rc = cam_i2c_write( cam_client->addr, 0x002E, 0x02, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x002E Write Err\n");
          return ERR_CODE;
        }

        rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x02, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity Err\n");
          return ERR_CODE;
        }

        rc = cam_i2c_read( cam_client->addr, 0x0011, &read_data, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_infinity NowMoniterMode Read Err\n");
            return ERR_CODE;
        }

        if(read_data == 0x00)
        {

            rc = cam_i2c_write( cam_client->addr, 0x0012, 0x01, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x0012 Err\n");
                return ERR_CODE;
            }
        }
        else
        {
            rc = cam_i2c_write( cam_client->addr, 0x0011, 0x00, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x0011 Write Err\n");
                return ERR_CODE;
            }
        }

        rc = cam006_check_internal_status( 0x02, 0x02 );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity mode change1 Err\n");
          return ERR_CODE;
        }

        rc = cam_i2c_write( cam_client->addr, 0x4852, 0x0000, WORD_LEN );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x4852 Err\n");
          return ERR_CODE;
        }

        rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x10, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity Err\n");
          return ERR_CODE;
        }

        rc = cam_i2c_write( cam_client->addr, 0x4850, 0x01, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x4850 Err\n");
          return ERR_CODE;
        }

        rc = cam006_check_internal_status( 0x10, 0x10 );
        if( rc < 0 )
        {
          ERR_LOG( "L_cam006_kernel_move_focus_infinity mode change2 Err\n");
          return ERR_CODE;
        }

        if(read_data != 0x00)
        {

            rc = cam_i2c_write( cam_client->addr, 0x0011, read_data, ADWORD_DATBYTE_LEN );
            if( rc < 0 )
            {
                ERR_LOG( "L_cam006_kernel_move_focus_infinity 0x0011 Write Err\n");
                return ERR_CODE;
            }

            rc = cam006_check_internal_status( 0x02, 0x03 );
            if( rc < 0 )
            {
                ERR_LOG( "L_cam006_kernel_move_focus_infinity mode change3 Err\n");
                return ERR_CODE;
            }
        }

    }

    return rc;
}

static int L_cam006_kernel_move_focus_optical_infinity(void)
{
    int rc = TRUE;
    uint16_t read_data;

    rc = cam_i2c_write( cam_client->addr, 0x4876, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x4876 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x002E, 0x02, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x002E Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x02, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x0011, &read_data, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x0011 Read Err\n");
        return ERR_CODE;
    }

    if(read_data == 0x00)
    {

        rc = cam_i2c_write( cam_client->addr, 0x0012, 0x01, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x0012 Write Err\n");
            return ERR_CODE;
        }
    }
    else
    {
        rc = cam_i2c_write( cam_client->addr, 0x0011, 0x00, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x0011 Write Err\n");
            return ERR_CODE;
        }
    }

    rc = cam006_check_internal_status( 0x02, 0x03 );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity mode change1 Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4852, cam_ctrl->cam006_af_otp_data.focus_optical_inf, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x4852 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x10, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4850, 0x01, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x4850 Write Err\n");
        return ERR_CODE;
    }

    rc = cam006_check_internal_status( 0x10, 0x10 );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity mode change2 Err\n");
        return ERR_CODE;
    }

    if(read_data != 0x00)
    {

        rc = cam_i2c_write( cam_client->addr, 0x0011, read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity 0x0011 Write Err\n");
            return ERR_CODE;
        }

        rc = cam006_check_internal_status( 0x02, 0x03 );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_optical_infinity mode change3 Err\n");
            return ERR_CODE;
        }
    }

    return rc;
}

int L_cam006_kernel_phyinf_to_optinf(void)
{
    int rc;
    uint16_t read_data;

    rc = cam_i2c_read( cam_client->addr, 0x6D7A, &read_data, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "L_cam006_kernel_phyinf_to_optinf 0x6D7A Read Err\n");
        return ERR_CODE;
    }

    if( read_data == 0 )
    {

        rc = L_cam006_kernel_move_focus_optical_infinity();
        if( rc < 0 )
        {
            ERR_LOG("L_cam006_kernel_phyinf_to_optinf Err\n");
            return ERR_CODE;
        }
    }
    return rc;
}

static int L_cam006_kernel_move_focus( uint16_t focus_pos )
{
    int rc = TRUE;

    rc = cam_i2c_write( cam_client->addr, 0x4876, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
    if( rc < 0 )
    {
      ERR_LOG( "L_cam006_kernel_move_focus 0x4876 Write Err\n");
      return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4852, focus_pos, WORD_LEN );
    if( rc < 0 )
    {
      ERR_LOG( "L_cam006_kernel_move_focus 0x4852 Write Err\n");
      return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x10, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
      ERR_LOG( "L_cam006_kernel_move_focus Err\n");
      return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4850, 0x01, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
      ERR_LOG( "L_cam006_kernel_move_focus 0x4850 Write Err\n");
      return ERR_CODE;
    }

    return rc;
}

static int L_cam006_kernel_flash_seq_auto_focus_ready( void )
{
    int rc = TRUE;
    unsigned long write_data;

    rc = cam_i2c_write( cam_client->addr, 0x002F, 0x00, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x002F Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x027D, 0x05, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x027D Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4066, 0x0F, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x4066 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x402F, 0x0B, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x402F Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x404C, 0x2C, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x404C Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x400B, 0x3C, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x400B Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x028C, 0x01, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x028C Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x0069, 0x0D, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x0069 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x02C4, 0x0010, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x02C4 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x0097, 0x02, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x0097 Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x0286, (uint16_t *)&cam_ctrl->flash_parm_info.pre_before_es, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x0286 Read Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_read( cam_client->addr, 0x028A, &cam_ctrl->flash_parm_info.pre_before_as, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x028A Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x026E, &cam_ctrl->flash_parm_info.pre_before_rt_r, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x026E Read Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_read( cam_client->addr, 0x0270, &cam_ctrl->flash_parm_info.pre_before_rt_b, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x0270 Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x16, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus INTCLR Err\n");
        return ERR_CODE;
    }

    rc = L_cam006_kernel_pre_flash();
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus L_cam006_kernel_pre_flash Err\n");
        return ERR_CODE;
    }

    cam_ctrl->pre_flash_on_flg = TRUE;

    rc = cam_i2c_write( cam_client->addr, 0x4876, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_start_auto_focus 0x4876 Write Err\n");
        return ERR_CODE;
    }

    if( cam_ctrl->af_mode == CAM_AF_MODE_AUTO )
    {
        write_data = 0x00;
    }
    else if( cam_ctrl->af_mode == CAM_AF_MODE_MACRO )
    {
        write_data = 0x02;
    }

    rc = cam_i2c_write( cam_client->addr, 0x01D3, write_data, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Auto Focus Search Set Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x0011, 0x01, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Change HalfRelease Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int L_cam006_kernel_flash_seq_auto_focus_off_ready( void )
{
    int rc = TRUE;

    rc = cam_i2c_write( cam_client->addr, 0x002F, 0x02, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_kernel_flash_seq_auto_focus_off_ready() Set HALFREL MANUAL MODE Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x027D, 0x05, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x027D Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4066, 0x0F, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x4066 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x402F, 0x0B, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x402F Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x404C, 0x2C, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x404C Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x400B, 0x3C, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x400B Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x028C, 0x01, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x028C Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x0069, 0x0D, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x0069 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x02C4, 0x0010, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x02C4 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x0097, 0x02, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x0097 Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x0286, (uint16_t *)&cam_ctrl->flash_parm_info.pre_before_es, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x0286 Read Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_read( cam_client->addr, 0x028A, &cam_ctrl->flash_parm_info.pre_before_as, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x028A Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x026E, &cam_ctrl->flash_parm_info.pre_before_rt_r, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x026E Read Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_read( cam_client->addr, 0x0270, &cam_ctrl->flash_parm_info.pre_before_rt_b, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Pre before 0x0270 Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x00FC, 0x16, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus INTCLR Err\n");
        return ERR_CODE;
    }

    rc = L_cam006_kernel_pre_flash();
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus L_cam006_kernel_pre_flash Err\n");
        return ERR_CODE;
    }

    cam_ctrl->pre_flash_on_flg = TRUE;

    rc = cam_i2c_write( cam_client->addr, 0x0011, 0x01, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Change HalfRelease Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int L_cam006_kernel_flash_threshold_check( int* flash_flg )
{
    int rc = TRUE;
    uint16_t as_now = 0;
    int32_t es_now = 0;

    if( cam_ctrl->flash_mode == CAM_FLASH_AUTO )
    {

        rc = cam_i2c_read( cam_client->addr, 0x0286, (uint16_t*)&es_now, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_flash_threshold_check 0x0286 Err\n");
            return ERR_CODE;
        }

        rc = cam_i2c_read( cam_client->addr, 0x028A, &as_now, WORD_LEN );
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_flash_threshold_check 0x028A Err\n");
            return ERR_CODE;
        }

        if( ((unsigned short)es_now) & 0x8000 )
        {
            es_now = - ((unsigned short)( ~es_now ) + 1);
        }

        if(( as_now + es_now ) < FLASH_THRESHOLD )
        {

            *flash_flg = TRUE;
        }
        else
        {

            *flash_flg = FALSE;
        }
    }
    else
    {
        *flash_flg = FALSE;
    }
    return rc;

}

static int L_cam006_kernel_pre_flash( void )
{
    int rc = TRUE;
    uint8_t data1;
    uint8_t data2;
    uint8_t data3;

    data1 = 0x73;

    data2 = 0x1E;

    data3 = 0xAA;

#ifndef CONFIG_TARGET_PROTOTYPE_WS1
    rc = cam_i2c_flash_write(( 0x82 >> 1 ), data1, data2, data3 );
    if( rc < 0 )
    {
        ERR_LOG("cam_i2c_flash_write Err\n");
        return ERR_CODE;
    }
#endif

    return rc;
}

static int L_cam006_kernel_main_flash( void )
{
    int rc = TRUE;
    uint8_t data1;
    uint8_t data2;
    uint8_t data3;

    data1 = 0x73;

    data2 = 0x1E;

    data3 = 0xAB;

#ifndef CONFIG_TARGET_PROTOTYPE_WS1
    rc = cam_i2c_flash_write(( 0x82 >> 1 ), data1, data2, data3 );
    if( rc < 0 )
    {
        ERR_LOG("cam_i2c_flash_write Err\n");
        return ERR_CODE;
    }
#endif

    return rc;
}

static int L_cam006_kernel_flash_off( void )
{
    int rc = TRUE;
    uint8_t data1;
    uint8_t data2;
    uint8_t data3;

    data1 = 0x73;

    data2 = 0x1E;

    data3 = 0xA8;

#ifndef CONFIG_TARGET_PROTOTYPE_WS1
    rc = cam_i2c_flash_write(( 0x82 >> 1 ), data1, data2, data3 );
    if( rc < 0 )
    {
        ERR_LOG("cam_i2c_flash_write Err\n");
        return ERR_CODE;
    }
#endif

    return rc;
}

static int L_cam006_kernel_still_flash_end( void )
{

    int rc = TRUE;

    rc = cam_i2c_write( cam_client->addr, 0x0069, 0x00, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "WB Proc OFF 0x0069 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x02C4, 0x00, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "WB Proc OFF 0x02C4 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x0097, 0x00, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "WB Proc OFF 0x0097 Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x445C, 0x00, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Vsync Fall AWB R Gain Offset Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x445E, 0x00, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Vsync Fall AWB B Gain Offset Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x028C, 0x00, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "Vsync Fall 0x028C Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x027D, 0x00, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "Vsync Fall 0x027D Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x0282, 0x0000, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Vsync Fall AE Gain Offset Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4066, 0x0A, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x4066 Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x402F, 0x0F, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x402F Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x404C, 0x20, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x404C Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x400B, 0x1A, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus 0x400B Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int L_cam006_kernel_set_main_flash_offset( void )
{
    int rc = TRUE;

    rc = cam_i2c_write( cam_client->addr, 0x0282, cam_ctrl->main_flash_cg_offset, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Vsync Fall AE Gain Offset Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x445C, cam_ctrl->main_flash_sc_shift_r, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Vsync Fall AWB R Gain Offset Err\n");
        return ERR_CODE;
    }
    rc = cam_i2c_write( cam_client->addr, 0x445E, cam_ctrl->main_flash_sc_shift_b, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Vsync Fall AWB B Gain Offset Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int L_cam006_kernel_check_flash_on( int* flash_onoff_flg )
{
    int rc = TRUE;
    int flash_check;
    int flash_flg;

    camera_flash_type flash_mode;

    if( flash_onoff_flg == NULL )
    {
        ERR_LOG("L_cam006_kernel_check_flash_on Parm NULL\n");
        return ERR_CODE;
    }

    if(( cam_ctrl->current_scene == CAM_SCENE_AUTO ) ||
       ( cam_ctrl->current_scene == CAM_SCENE_PORTRAIT ) ||
       ( cam_ctrl->current_scene == CAM_SCENE_SCENERY ))
    {
        flash_mode = cam_ctrl->flash_mode;
    }

    else
    {
        flash_mode = scene_info[ cam_ctrl->current_scene ].flash;
    }

    if( flash_mode == CAM_FLASH_ON )
    {
        flash_flg = TRUE;
    }

    else if( flash_mode == CAM_FLASH_OFF )
    {
        flash_flg = FALSE;
    }

    else
    {

        rc = L_cam006_kernel_flash_threshold_check( &flash_check );
        if( rc < 0 )
        {
            ERR_LOG("L_cam006_kernel_flash_threshold_check() Err\n");
            return ERR_CODE;
        }

        if( flash_check == TRUE )
        {

            flash_flg = TRUE;
        }
        else
        {

            flash_flg = FALSE;
        }
    }
    *flash_onoff_flg = flash_flg;

    return rc;
}

static int L_cam006_kernel_calc_capture_wait_time( int32_t* wait_time )
{
    int rc = TRUE;
    uint16_t vmax_l;
    uint16_t vmax_h;
    uint16_t hmax_l;
    uint16_t hmax_h;
    uint32_t vsync_time;

    if( wait_time == NULL )
    {
        ERR_LOG("L_cam006_kernel_calc_capture_wait_time Parm NULL Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x3603, &vmax_l, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x3603 Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x3604, &vmax_h, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x3604 Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x3605, &hmax_l, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x3605 Read Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x3606, &hmax_h, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x3606 Read Err\n");
        return ERR_CODE;
    }

    vsync_time = (uint32_t)(( (( hmax_h * 256 ) + hmax_l + 1 ) * 2 * ( ( vmax_h * 256 ) + vmax_l + 1 ) ) / 43200);

    vsync_time += 1;

    *wait_time = vsync_time - DELAY_MARGIN;

    if( *wait_time < 0 )
    {
        *wait_time = 0;
    }

    return rc;

}

static int L_cam006_kernel_get_shutter_speed(void)
{
    int rc = TRUE;
    uint16_t read_data;

    rc = cam_i2c_read( cam_client->addr, 0x00F4, &read_data, WORD_LEN);
    if( rc < 0 )
    {
        ERR_LOG("0x00F4 Read Err\n");
        cam_ctrl->exif_exposure_time = 0xFFFFFFFF;
        return ERR_CODE;
    }

    cam_ctrl->exif_exposure_time = 0;

    cam_ctrl->exif_exposure_time = (uint32_t)read_data;

    rc = cam_i2c_read( cam_client->addr, 0x00F2, &read_data, WORD_LEN);
    if( rc < 0 )
    {
        ERR_LOG("0x00F2 Read Err\n");
        cam_ctrl->exif_exposure_time = 0xFFFFFFFF;
        return ERR_CODE;
    }

    cam_ctrl->exif_exposure_time <<= 16;
    cam_ctrl->exif_exposure_time |= (uint32_t)read_data;

    return rc;
}

static int L_cam006_kernel_get_now_sensor_mode( uint16_t* current_mode )
{
    int rc = TRUE;
    uint16_t read_data;

    if( current_mode == NULL )
    {
        ERR_LOG("L_cam006_kernel_get_now_sensor_mode Parm NULL Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_read( cam_client->addr, 0x0004, &read_data, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG("Get Now Sensor Mode Err\n");
        return ERR_CODE;
    }

    *current_mode = read_data;

    return rc;

}

static int L_cam006_kernel_set_v_fall_proc( camera_vsync_fall_irq_type v_fall_proc )
{
    int i;

    for( i=0; i < V_FALL_PROC_RESERVE_MAX; i++ )
    {
        if( cam006_v_fall_proc_reserve_tbl[i] == CAMERA_V_FALL_NONE )
        {

            break;
        }
    }

    if( i >= V_FALL_PROC_RESERVE_MAX )
    {
        ERR_LOG("V FALL RESERVE TBL FULL\n");
        return ERR_CODE;
    }

    cam006_v_fall_proc_reserve_tbl[i] = v_fall_proc;

    return TRUE;
}

static int L_cam006_kernel_get_v_fall_proc( camera_vsync_fall_irq_type* v_fall_proc_reserve )
{
    int i;

    if( v_fall_proc_reserve == NULL )
    {
        ERR_LOG("L_cam006_kernel_get_v_fall_proc Parm NULL Err\n");
        return ERR_CODE;
    }

    for( i=0; i < V_FALL_PROC_RESERVE_MAX; i++ )
    {
        if( cam006_v_fall_proc_reserve_tbl[i] != CAMERA_V_FALL_NONE )
        {

            *v_fall_proc_reserve = cam006_v_fall_proc_reserve_tbl[i];
            break;
        }
    }

    if( i >= V_FALL_PROC_RESERVE_MAX )
    {

        *v_fall_proc_reserve = CAMERA_V_FALL_NONE;
    }
    else
    {

        for( ++i; i < V_FALL_PROC_RESERVE_MAX; i++ )
        {
            if( i >= ( V_FALL_PROC_RESERVE_MAX - 1 ))
            {
                cam006_v_fall_proc_reserve_tbl[i] = CAMERA_V_FALL_NONE;
            }
            else
            {
                cam006_v_fall_proc_reserve_tbl[i-1] = cam006_v_fall_proc_reserve_tbl[i];
            }
        }
    }

    return TRUE;

}

static int L_cam006_kernel_get_vfall_proc_table( camera_vsync_fall_irq_type* v_fall_proc_reserve )
{

    if( v_fall_proc_reserve == NULL )
    {
        ERR_LOG("L_cam006_kernel_get_vfall_proc_table Parm NULL Err\n");
        return ERR_CODE;
    }

    *v_fall_proc_reserve = cam006_v_fall_proc_reserve_tbl[0];

    return TRUE;
}

static void L_cam006_kernel_set_vfall_free_irq( void )
{
    camera_vsync_fall_irq_type get_vfall_tbl;
    int rc;

    rc = L_cam006_kernel_get_vfall_proc_table( &get_vfall_tbl );

    L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_NONE );

}

static int cam_set_sensor_mode(int mode, struct sensor_pict_fps *gfps)
{
  long   rc = 0;

  switch(mode) {
    case SENSOR_PREVIEW_MODE:

        if( cam_ctrl->main_flash_on_flg == TRUE )
        {

            rc = L_cam006_kernel_flash_off();
            if( rc < 0 )
            {
                ERR_LOG("cam006_still_to_preview flash off\n");
                return ERR_CODE;
            }

            cam_ctrl->main_flash_on_flg = FALSE;
        }

        L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_NONE );
        L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

        if( cam_ctrl->picture_size_chenge != TRUE )
        {

            return TRUE;
        }

        rc = L_cam006_start_preview();
        if( rc < 0 )
        {
          ERR_LOG("L_cam006_start_preview Err\n");
          return rc;
        }

        cam_ctrl->camera_state = CAMERA_PREVIEW_STATE;

        cam_ctrl->picture_size_chenge = FALSE;

        break;

    case SENSOR_SNAPSHOT_MODE:

        cam_ctrl->still_mode = STILL_YUV_MODE;

        rc = L_cam006_kernel_phyinf_to_optinf();
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_infinity NowFocusPos Read Err\n");
            return ERR_CODE;
        }

#ifndef CONFIG_TARGET_PROTOTYPE_WS1

        cam_ctrl->flash_on_flg = cam_ctrl->flash_parm_info.flash_flg;

        if( cam_ctrl->flash_parm_info.flash_flg == TRUE )
        {

            cam_ctrl->still_err_flg = FALSE;

            cam_ctrl->wait_flag = 0;

            free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

            L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

            rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_high_interrupt, IRQF_TRIGGER_RISING, "cam_vsync_h", 0);
            if (rc < 0)
            {
                ERR_LOG("cam vsync high request_irq failed!\n");
            }

            if( cam_ctrl->af_proc_call_flg == TRUE )
            {

                L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_STILL_MAIN_FLASH_AF_ON );
            }

            else
            {

                L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_MAIN_FLASH_AF_OFF );

            }

            rc = cam_wait_event();
            if( rc < 0 )
            {
                ERR_LOG("SENSOR_SNAPSHOT_MODE Current Thread Wait Err\n");
                return ERR_CODE;
            }

            if( cam_ctrl->still_err_flg == TRUE )
            {
                return ERR_CODE;
            }
            else
            {
                rc = TRUE;
            }
        }

        else
        {

            rc = L_cam006_kernel_get_shutter_speed();
            if( rc < 0 )
            {
                ERR_LOG("YUV STILL FLASH OFF Get ShutterSpeed Err\n");
                return ERR_CODE;
            }
        }

        cam_ctrl->af_proc_call_flg = FALSE;

#endif

        cam_ctrl->camera_state = CAMERA_STILL_STATE;

        break;

    case SENSOR_RAW_SNAPSHOT_MODE:

        cam_ctrl->still_mode = STILL_JPEG_MODE;

        rc = L_cam006_kernel_phyinf_to_optinf();
        if( rc < 0 )
        {
            ERR_LOG( "L_cam006_kernel_move_focus_infinity NowFocusPos Read Err\n");
            return ERR_CODE;
        }

        cam_ctrl->flash_on_flg = cam_ctrl->flash_parm_info.flash_flg;

#ifndef CONFIG_TARGET_PROTOTYPE_WS1
        if( cam_ctrl->flash_parm_info.flash_flg == TRUE )
        {

            cam_ctrl->main_flash_off_v_cnt = 0;
            cam_ctrl->yuv_main_flash_on_v_cnt = 0;

            cam_ctrl->still_err_flg = FALSE;

            cam_ctrl->wait_flag = 0;

            free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

            L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

            rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_high_interrupt, IRQF_TRIGGER_RISING, "cam_vsync_h", 0);
            if (rc < 0)
            {
                ERR_LOG("cam vsync high request_irq failed!\n");
                return ERR_CODE;
            }

            if( cam_ctrl->af_proc_call_flg == TRUE )
            {
                L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_JPEG_STILL_MAIN_FLASH_AF_ON );
            }

            else
            {

                L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_MAIN_FLASH_AF_OFF );

            }

            rc = cam_wait_event();
            if( rc < 0 )
            {
                ERR_LOG("SENSOR_SNAPSHOT_MODE Current Thread Wait Err\n");
                return ERR_CODE;
            }

            if( cam_ctrl->still_err_flg == TRUE )
            {
                return ERR_CODE;
            }
            else
            {
                rc = TRUE;
            }
        }
        else
#endif

        {

            cam_ctrl->main_flash_off_v_cnt = 0;
            cam_ctrl->yuv_main_flash_on_v_cnt = 0;
            cam_ctrl->normal_still_v_cnt = 0;

            cam_ctrl->still_err_flg = FALSE;

            cam_ctrl->wait_flag = 0;

            free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

            L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

            rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_high_interrupt, IRQF_TRIGGER_RISING, "cam_vsync_h", 0);
            if (rc < 0)
            {
                ERR_LOG("cam vsync high request_irq failed!\n");
                return ERR_CODE;
            }

            L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_JPEG_STILL_NORMAL );

            rc = cam_wait_event();
            if( rc < 0 )
            {
                ERR_LOG("SENSOR_SNAPSHOT_MODE Current Thread Wait Err\n");
                return ERR_CODE;
            }

            if( cam_ctrl->still_err_flg == TRUE )
            {
                return ERR_CODE;
            }
            else
            {
                rc = TRUE;
            }
        }

        cam_ctrl->af_proc_call_flg = FALSE;

        cam_ctrl->camera_state = CAMERA_STILL_STATE;

        break;
    case SENSOR_DLC_MODE:

        L_cam006_kernel_sensor_power_off();

        mdelay(100);

        rc = L_cam006_kernel_sensor_dlc_power_on();
        if( rc < 0 )
        {
            ERR_LOG("SENSOR_DLC_MODE L_cam006_kernel_sensor_power_on Err\n");
            return ERR_CODE;
        }

        break;

    default:
        
        rc = -EFAULT;
        break;
  }

  return rc;
}

static int32_t cam_set_move_focus(int af_mode)
{

  cam_ctrl->af_cmd.cmd           = SINGLE_CMD;

  if(af_mode == AF_SINGLE_MODE)
    cam_ctrl->af_cmd.param       = 0x1002;
  else
    cam_ctrl->af_cmd.param       = 0x1042;

  if(cam_ctrl->flash_mode == FLASH_STILL_MODE) {

    cam_vsync_wait(PRE_FLASH_ON_VREF_INT,VSYNC_HIGH);
  }
  else {

    cam_chg_state(SENSOR_STATE_AF, SENSOR_AF_SEND_CMD);
    cam_vsync_wait(AF_VREF_INT,VSYNC_HIGH);
  }

  return 0;
}

static int32_t cam_set_default_focus(void)
{
    int rc = TRUE;

    cam_ctrl->wait_flag = 0;

    if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_NONE )
    {

        free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

        rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0);
        if (rc < 0)
        {
            ERR_LOG("cam vsync low request_irq failed!\n");
            return ERR_CODE;
        }

        L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_DEFAULT_FOCUS );
    }
    else
    {

        rc = L_cam006_kernel_set_v_fall_proc( CAMERA_V_FALL_DEFAULT_FOCUS );
        if( rc < 0 )
        {
            ERR_LOG("SET V FALL TBL CAMERA_V_FALL_DEFAULT_FOCUS ERR\n");
          return ERR_CODE;
        }
    }

    rc = cam_wait_event();
    if( rc < 0 )
    {
        ERR_LOG("cam006_still_to_preview Current Thread Wait Err\n");
        return ERR_CODE;
    }

    return rc;
}

static int cam_init_client(struct i2c_client *client)
{

    init_waitqueue_head(&cam_wait_queue);

    return 0;
}

static void L_cam006_entry_vsync_low_proc( camera_vsync_fall_irq_type entry )
{

    camera_vsync_fall_irq_type v_fall_proc_reserve;
    int rc;

    rc = L_cam006_kernel_get_v_fall_proc( &v_fall_proc_reserve );
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_kernel_get_v_fall_proc Get V Fall Reserve Proc Err\n");
        v_fall_proc_reserve = CAMERA_V_FALL_NONE;
    }

    if( entry == CAMERA_V_FALL_NONE )
    {
        if( v_fall_proc_reserve != CAMERA_V_FALL_NONE )
        {
            entry = v_fall_proc_reserve;
        }
    }

    cam_ctrl->v_fall_proc = entry;
}

static void L_cam006_entry_vsync_high_proc( camera_vsync_rise_irq_type entry )
{
    cam_ctrl->v_rise_proc = entry;
}

static void L_cam006_delete_vsync_high( void )
{
    L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

}

static int cam006_kernel_set_dimensions( struct sensor_cfg_data cfg_data )
{
  uint16_t set_picture_width = cfg_data.sensor_size.picture_width;
  uint16_t set_picture_height = cfg_data.sensor_size.picture_height;

  uint16_t set_display_width = cfg_data.sensor_size.display_width;
  uint16_t set_display_height = cfg_data.sensor_size.display_height;
  uint16_t set_fps = cfg_data.cfg.gfps.prevfps;

  if( cam_ctrl->camera_state == CAMERA_INIT_STATE )
  {
    ERR_LOG( "Camera State Err\n");
    return ERR_CODE;
  }

  if((( set_picture_width == cam_ctrl->current_picture_width )    &&
      ( set_picture_height == cam_ctrl->current_picture_height )) &&
      (( set_display_width == cam_ctrl->current_display_width ) &&
       ( set_display_height == cam_ctrl->current_display_height )) &&
       ( set_fps == cam_ctrl->current_fps ))
  {
    cam_ctrl->picture_size_chenge = FALSE;

    return TRUE;
  }

  cam_ctrl->picture_size_chenge = TRUE;

  cam_ctrl->past_picture_width = cam_ctrl->current_picture_width;
  cam_ctrl->past_picture_height = cam_ctrl->current_picture_height;
  cam_ctrl->past_display_width = cam_ctrl->current_display_width;
  cam_ctrl->past_display_height = cam_ctrl->current_display_height;

  cam_ctrl->current_picture_width = set_picture_width;
  cam_ctrl->current_picture_height = set_picture_height;
  cam_ctrl->current_display_width = set_display_width;
  cam_ctrl->current_display_height = set_display_height;

  cam_ctrl->past_fps = cam_ctrl->current_fps;

  cam_ctrl->current_fps = set_fps;

  return TRUE;
}

static int cam006_still_to_preview( void )
{
  int rc = 0;

  cam_ctrl->wait_flag = 0;

  if( cam_ctrl->v_fall_proc == CAMERA_V_FALL_NONE )
  {

      free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

      rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0);
      if (rc < 0)
      {
          ERR_LOG("cam vsync low request_irq failed!\n");
         return ERR_CODE;
      }

      L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE );
  }
  else
  {

      rc = L_cam006_kernel_set_v_fall_proc( CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE );
      if( rc < 0 )
      {
          ERR_LOG("SET V FALL TBL CAMERA_V_FALL_STILL_TO_CHANGE_PREVIEW_PROC_RESERVE ERR\n");
        return ERR_CODE;
      }
  }

  rc = cam_wait_event();
  if( rc < 0 )
  {
      ERR_LOG("cam006_still_to_preview Current Thread Wait Err\n");
      return ERR_CODE;
  }

  cam_ctrl->camera_state = CAMERA_PREVIEW_STATE;

  return rc;

}

static int cam006_start_auto_focus( uint8_t focus_mode )
{
  int rc = TRUE;

  if( cam_ctrl->camera_state != CAMERA_PREVIEW_STATE )
  {
    ERR_LOG( "cam006_start_auto_focus Camera State Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->focus_state == CAMERA_AF_FOCUS ) ||
     ( cam_ctrl->focus_state == CAMERA_MOVE_FOCUS ))
  {
    ERR_LOG( "cam006_start_auto_focus FocusState Err\n");
    return ERR_CODE;
  }

  rc = L_cam006_kernel_move_focus_infinity();
  if( rc < 0 )
  {
    ERR_LOG( "cam006_start_auto_focus L_cam006_kernel_move_focus_infinity Err\n");
    return ERR_CODE;
  }

  if( focus_mode == CAM_AF_MODE_NORMAL )
  {

    cam_ctrl->af_mode = focus_mode;
    return TRUE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x4876, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG( "cam006_start_auto_focus Search Area LOW Write Err\n");
    return ERR_CODE;
  }

  if( focus_mode == CAM_AF_MODE_AUTO )
  {

    rc = cam_i2c_write_table( &cam_regs.af_inf_to_macro_start[0],
                                   cam_regs.af_inf_to_macro_start_reg_setting_size );
    if( rc < 0 )
    {
      ERR_LOG( "cam006_start_auto_focus inf to macro Err\n");
      return ERR_CODE;
    }
  }
  else if( focus_mode == CAM_AF_MODE_MACRO )
  {

    rc = cam_i2c_write_table( &cam_regs.af_macro_to_inf_start[0],
                                   cam_regs.af_macro_to_inf_start_reg_setting_size );
    if( rc < 0 )
    {
      ERR_LOG( "cam006_start_auto_focus macro to inf Err\n");
      return ERR_CODE;
    }
  }

  free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

  cam_ctrl->af_mode = focus_mode;
  cam_ctrl->focus_state = CAMERA_AF_FOCUS;

  cam_ctrl->wait_flag = 0;

  if (request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0))
  {
    ERR_LOG("Start Auto Focus Vsycn Fall Trig Set Err\n");
    cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;
    return ERR_CODE;
  }

  L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_AF_MODE );

  rc = cam_wait_event();
  if( rc < 0 )
  {
    ERR_LOG("Start Auto Focus Current Thread Wait Err\n");
    cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;
    return ERR_CODE;
  }

  if( cam_ctrl->af_result == TRUE )
  {
    rc = TRUE;
  }
  else
  {
    rc = ERR_CODE;
  }

  cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;

  return rc;

}

static int cam006_kernel_flash_seq_auto_focus( camera_af_mode_type focus_mode )
{
    int rc = TRUE;

    if( cam_ctrl->camera_state != CAMERA_PREVIEW_STATE )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus Camera State Err\n");
        return ERR_CODE;
    }

    if(( cam_ctrl->focus_state == CAMERA_AF_FOCUS ) ||
     ( cam_ctrl->focus_state == CAMERA_MOVE_FOCUS ))
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus FocusState Err\n");
        return ERR_CODE;
    }

    rc = L_cam006_kernel_move_focus_infinity();
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_flash_seq_auto_focus L_cam006_kernel_move_focus_infinity Err\n");
        return ERR_CODE;
    }

    if( focus_mode == CAM_AF_MODE_NORMAL )
    {

        cam_ctrl->af_mode = focus_mode;
        return TRUE;
    }

    free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

    cam_ctrl->af_mode = focus_mode;
    cam_ctrl->focus_state = CAMERA_AF_FOCUS;

    cam_ctrl->wait_flag = 0;

    if (request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_low_interrupt, IRQF_TRIGGER_FALLING, "cam_vsync_l", 0))
    {
        ERR_LOG("cam006_kernel_flash_seq_auto_focus Vsycn Fall Trig Set Err\n");
        cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;
        return ERR_CODE;
    }

    L_cam006_entry_vsync_low_proc( CAMERA_V_FALL_AF_PRE_FLASH );

    rc = cam_wait_event();
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_flash_seq_auto_focus Current Thread Wait Err\n");
        cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;
        return ERR_CODE;
    }

    if( cam_ctrl->af_result == TRUE )
    {
        rc = TRUE;
    }
    else
    {
        rc = ERR_CODE;
    }

    cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;

    return rc;

}

static int cam006_kernel_set_wb( camera_wb_type wb )
{
  int rc = TRUE;
  unsigned char write_data;

  if( wb >= CAM_WB_MAX )
  {
    ERR_LOG("cam006_kernel_set_wb Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_wb == wb ) &&
     ( cam_ctrl->internal_power_offon_retry_flg == FALSE ) &&
     ( cam_ctrl->current_wb_update == FALSE ))
  {

    return TRUE;
  }

  if( cam_ctrl->current_scene != CAM_SCENE_AUTO)
  {
    cam_ctrl->current_wb = wb;
    cam_ctrl->current_wb_update = TRUE;

    return TRUE;
  }

  switch( wb )
  {
    case CAM_WB_AUTO:

      write_data = 0x20;
      break;
    case CAM_WB_INCANDESCENT:

      write_data = 0x08;
      break;
    case CAM_WB_FLUORESCENT:

      write_data = 0x07;
      break;
    case CAM_WB_DAYLIGHT:

      write_data = 0x04;
      break;
    case CAM_WB_CLOUDY_DAYLIGHT:

      write_data = 0x06;
      break;
    default:
      ERR_LOG("cam006_kernel_set_wb Parm Err\n");
      return ERR_CODE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x0102, (unsigned long)write_data, ADWORD_DATBYTE_LEN);
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_wb Reg Write Err\n");
    return ERR_CODE;
  }

  cam_ctrl->current_wb = wb;
  cam_ctrl->current_wb_update = FALSE;

  return rc;
}

static int cam006_kernel_set_effect( camera_effect_type effect )
{
  int rc = TRUE;
  unsigned char write_data;

  if( effect >= CAM_EFFECT_MAX )
  {
    ERR_LOG("cam006_kernel_set_effect Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_effect == effect ) && ( cam_ctrl->internal_power_offon_retry_flg == FALSE ))
  {

    return TRUE;
  }

  switch( effect )
  {
    case CAM_EFFECT_NONE:

      write_data = 0x00;
      break;
    case CAM_EFFECT_MONO:

      write_data = 0x04;
      break;
    case CAM_EFFECT_SEPIA:

      write_data = 0x03;
      break;
    case CAM_EFFECT_NEGA:

      write_data = 0x02;
      break;
    case CAM_EFFECT_SOLARI:

      write_data = 0x01;
      break;
    default:
      ERR_LOG("cam006_kernel_set_effect Parm Err\n");
      return ERR_CODE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x005F, (unsigned long)write_data, ADWORD_DATBYTE_LEN);
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_effect Reg Write Err\n");
    return FALSE;
  }

  cam_ctrl->current_effect = effect;

  return rc;
}

static int cam006_kernel_set_iso( camera_iso_type iso )
{
  int rc = TRUE;
  unsigned char write_data;

  if( iso >= CAM_ISO_MAX )
  {
    ERR_LOG("cam006_kernel_set_iso Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_iso == iso ) &&
     ( cam_ctrl->internal_power_offon_retry_flg == FALSE ) &&
     ( cam_ctrl->current_iso_update == FALSE ))
  {

    return TRUE;
  }

  if( cam_ctrl->current_scene != CAM_SCENE_AUTO)
  {
    cam_ctrl->current_iso = iso;
    cam_ctrl->current_iso_update = TRUE;

    return TRUE;
  }

  switch( iso )
  {
    case CAM_ISO_AUTO:

      write_data = 0x00;
      break;
    case CAM_ISO_100:

      write_data = 0x07;
      break;
    case CAM_ISO_200:

      write_data = 0x0A;
      break;
    case CAM_ISO_400:

      write_data = 0x0D;
      break;
    default:
      ERR_LOG("cam006_kernel_set_iso Parm Err\n");
      return ERR_CODE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x01E5, (unsigned long)write_data, ADWORD_DATBYTE_LEN);
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_iso Reg Write Err\n");
    return ERR_CODE;
  }

  cam_ctrl->current_iso = iso;
  cam_ctrl->current_iso_update = FALSE;

  return rc;

}

static int cam006_kernel_set_ae_mode( camera_auto_exposure_type aec_mode )
{
  int rc = TRUE;
  unsigned char write_data;

  if( aec_mode >= CAM_AEC_MAX )
  {
    ERR_LOG("cam006_kernel_set_ae_mode Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_ae_mode == aec_mode ) &&
     ( cam_ctrl->internal_power_offon_retry_flg == FALSE ) &&
     ( cam_ctrl->current_ae_mode_update == FALSE ))
  {

    return TRUE;
  }

  if( cam_ctrl->current_scene != CAM_SCENE_AUTO)
  {
    cam_ctrl->current_ae_mode = aec_mode;
    cam_ctrl->current_ae_mode_update = TRUE;

    return TRUE;
  }

  switch( aec_mode )
  {
    case CAM_AEC_AVERAGE:

      write_data = 0x00;
      break;
    case CAM_AEC_CENTER:

      write_data = 0x01;
      break;
    case CAM_AEC_SPOT:

      write_data = 0x02;
      break;
    default:
      ERR_LOG("cam006_kernel_set_ae_mode Parm Err\n");
      return ERR_CODE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x0104, (unsigned long)write_data, ADWORD_DATBYTE_LEN);
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_ae_mode Reg Write Err\n");
    return ERR_CODE;
  }

  cam_ctrl->current_ae_mode = aec_mode;
  cam_ctrl->current_ae_mode_update = FALSE;

  return rc;

}

static int cam006_kernel_set_af_window( struct afwin_info_cfg *afwin_info )
{
  int rc = TRUE;

  if(afwin_info->width < AF_WINDOW_WIDTH_SIZE_MIN ||
     afwin_info->width > AF_WINDOW_WIDTH_SIZE_MAX ||
     afwin_info->height < AF_WINDOW_HEIGHT_SIZE_MIN ||
     afwin_info->height > AF_WINDOW_HEIGHT_SIZE_MAX ||
     afwin_info->pos_x < AF_WINDOW_OFFSET_X ||
     afwin_info->pos_y < AF_WINDOW_OFFSET_Y ||
     afwin_info->pos_x + afwin_info->width > DEFAULT_STILL_5M_WIDTH + AF_WINDOW_OFFSET_X - AF_WINDOW_UNS_X ||
     afwin_info->pos_y + afwin_info->height> DEFAULT_STILL_5M_HEIGHT+ AF_WINDOW_OFFSET_Y - AF_WINDOW_UNS_Y)
  {
    ERR_LOG("cam006_kernel_set_af_window Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_afwin_info.width  == afwin_info->width &&
     cam_ctrl->current_afwin_info.height == afwin_info->height &&
     cam_ctrl->current_afwin_info.pos_x  == afwin_info->pos_x &&
     cam_ctrl->current_afwin_info.pos_y  == afwin_info->pos_y ) &&
     ( cam_ctrl->internal_power_offon_retry_flg == FALSE ))
  {

    return TRUE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x4c4c, (unsigned long)afwin_info->pos_x, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_af_window Reg1 Write Err\n");
    return ERR_CODE;
  }
  rc = cam_i2c_write( cam_client->addr, 0x4c4e, (unsigned long)afwin_info->pos_y, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_af_window Reg2 Write Err\n");
    return ERR_CODE;
  }
  rc = cam_i2c_write( cam_client->addr, 0x4c50, (unsigned long)afwin_info->width, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_af_window Reg3 Write Err\n");
    return ERR_CODE;
  }
  rc = cam_i2c_write( cam_client->addr, 0x4c52, (unsigned long)afwin_info->height, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_af_window Reg4 Write Err\n");
    return ERR_CODE;
  }

  cam_ctrl->current_afwin_info.width      = afwin_info->width;
  cam_ctrl->current_afwin_info.height     = afwin_info->height;
  cam_ctrl->current_afwin_info.pos_x      = afwin_info->pos_x;
  cam_ctrl->current_afwin_info.pos_y      = afwin_info->pos_y;

  return rc;
}

static int cam006_kernel_set_ae_window( struct aeawb_win_info_cfg *aewin_info )
{
  int rc = TRUE;

  if(aewin_info->width < AE_WINDOW_WIDTH_MIN ||
     aewin_info->width > AE_WINDOW_WIDTH_MAX ||
     aewin_info->height < AE_WINDOW_HEIGHT_MIN ||
     aewin_info->height > AE_WINDOW_HEIGHT_MAX)
  {
    ERR_LOG("cam006_kernel_set_ae_window Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_aewin_info.width  == aewin_info->width    &&
       cam_ctrl->current_aewin_info.height == aewin_info->height ) &&
     ( cam_ctrl->internal_power_offon_retry_flg == FALSE ))
  {

    return TRUE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x4c02, (unsigned long)aewin_info->width, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_ae_window Reg1 Write Err\n");
    return ERR_CODE;
  }
  rc = cam_i2c_write( cam_client->addr, 0x4c04, (unsigned long)aewin_info->height, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_ae_window Reg2 Write Err\n");
    return ERR_CODE;
  }

  cam_ctrl->current_aewin_info.width  = aewin_info->width;
  cam_ctrl->current_aewin_info.height = aewin_info->height;

  return rc;
}

static int cam006_kernel_set_awb_window( struct aeawb_win_info_cfg *awbwin_info )
{
  int rc = TRUE;

  if(awbwin_info->width < AWB_WINDOW_WIDTH_MIN ||
     awbwin_info->width > AWB_WINDOW_WIDTH_MAX ||
     awbwin_info->height < AWB_WINDOW_HEIGHT_MIN ||
     awbwin_info->height > AWB_WINDOW_HEIGHT_MAX)
  {
    ERR_LOG("cam006_kernel_set_awb_window Parm Err\n");
    return ERR_CODE;
  }

  if(( cam_ctrl->current_awbwin_info.width  == awbwin_info->width    &&
       cam_ctrl->current_awbwin_info.height == awbwin_info->height ) &&
     ( cam_ctrl->internal_power_offon_retry_flg == FALSE ))
  {

    return TRUE;
  }

  rc = cam_i2c_write( cam_client->addr, 0x4c0a, (unsigned long)awbwin_info->width, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_ae_window Reg1 Write Err\n");
    return ERR_CODE;
  }
  rc = cam_i2c_write( cam_client->addr, 0x4c0c, (unsigned long)awbwin_info->height, WORD_LEN );
  if( rc < 0 )
  {
    ERR_LOG("cam006_kernel_set_ae_window Reg2 Write Err\n");
    return ERR_CODE;
  }

  cam_ctrl->current_awbwin_info.width  = awbwin_info->width;
  cam_ctrl->current_awbwin_info.height = awbwin_info->height;

  return rc;
}

static int cam006_kernel_set_lens_shading( uint8_t shading_flg )
{
    int rc = TRUE;
    uint16_t write_data;

    if(( cam_ctrl->current_lens_shading == shading_flg ) && ( cam_ctrl->internal_power_offon_retry_flg == FALSE ))
    {

        return TRUE;
    }

    if( shading_flg == TRUE )
    {

        write_data = 0x5F;
    }
    else
    {

        write_data = 0x50;
    }

    rc = cam_i2c_write( cam_client->addr, 0x0068, write_data, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        ERR_LOG( "cam006_kernel_set_lens_shading I2C Write Err\n");
        return ERR_CODE;
    }

    cam_ctrl->current_lens_shading = shading_flg;

    return rc;

}

static void cam006_kernel_load_sceneval( struct cam_i2c_reg_conf* dst, struct cam_i2c_reg_conf* src, int num )
{
  int pos ;
  for( pos = 0 ; pos < num ; pos++)
  {
    *dst = *src ;
    dst++ ;
    src++ ;
  }
}

static void cam006_kernel_update_sceneval( struct cam_i2c_reg_conf* dst, int num, unsigned short waddr, int val )
{
  int pos ;
  unsigned long wdata ;

  switch(waddr)
  {

  case 0x0102:
    switch( ( camera_wb_type ) val )
    {
      case CAM_WB_AUTO:
        wdata = 0x20;
        break;
      case CAM_WB_INCANDESCENT:
        wdata = 0x08;
        break;
      case CAM_WB_FLUORESCENT:
        wdata = 0x07;
        break;
      case CAM_WB_DAYLIGHT:
        wdata = 0x04;
        break;
      case CAM_WB_CLOUDY_DAYLIGHT:
        wdata = 0x06;
        break;
      default:
        wdata = 0x20;
        break;
    }
    break ;

  case 0x0104:
    switch( ( camera_auto_exposure_type ) val )
    {
      case CAM_AEC_AVERAGE:
        wdata = 0x00;
        break;
      case CAM_AEC_CENTER:
        wdata = 0x01;
        break;
      case CAM_AEC_SPOT:
        wdata = 0x02;
        break;
      default:
        wdata = 0x00;
        break;
    }
    break ;

  case 0x01E5:
    switch( ( camera_iso_type ) val )
    {
      case CAM_ISO_AUTO:
        wdata = 0x00;
        break;
      case CAM_ISO_100:
        wdata = 0x07;
        break;
      case CAM_ISO_200:
        wdata = 0x0A;
        break;
      case CAM_ISO_400:
        wdata = 0x0D;
        break;
      default:
        wdata = 0x00;
        break;
    }
    break ;
  default :
    return ;
  }

  for( pos = 0 ; pos < num ; pos++)
  {
    if(dst->waddr == waddr)
    {
      dst->wdata = wdata ;
      break ;
    }
    dst++ ;
  }
}

static int cam006_kernel_set_scene( camera_scene_type scene_mode )
{
    int rc = TRUE;

    int set_num;
    struct cam_i2c_reg_conf set_scene[30];
    uint8_t save_wb_update;
    uint8_t save_iso_update;
    uint8_t save_ae_mode_update;
    camera_scene_type save_scene_mode;

    if( scene_mode >= CAM_SCENE_MAX )
    {
        ERR_LOG("cam006_kernel_set_scene Parm Err:%d\n", scene_mode);
        return ERR_CODE;
    }

    if( scene_mode == CAM_SCENE_AUTO )
    {
      if(( cam_ctrl->current_wb_update == FALSE ) &&
         ( cam_ctrl->current_ae_mode_update == FALSE ) &&
         ( cam_ctrl->current_iso_update == FALSE ))
      {
        if(( cam_ctrl->current_scene == scene_mode ) &&
           ( cam_ctrl->current_scene_update == FALSE ))
        {

            return TRUE;
        }
      }
    }
    else
    {
      if(( cam_ctrl->current_scene == scene_mode ) &&
         ( cam_ctrl->current_scene_update == FALSE ))
      {

          return TRUE;
      }
    }

    save_wb_update = cam_ctrl->current_wb_update ;
    save_iso_update = cam_ctrl->current_iso_update ;
    save_ae_mode_update = cam_ctrl->current_ae_mode_update ;

    switch( scene_mode )
    {
        case CAM_SCENE_AUTO:

            set_num = cam_regs.set_scene_auto_reg_size ;
            cam006_kernel_load_sceneval( &set_scene[0], &cam_regs.set_scene_auto_reg[0], set_num) ;

            cam006_kernel_update_sceneval(&set_scene[0], set_num, 0x0102, cam_ctrl->current_wb) ;
            cam006_kernel_update_sceneval(&set_scene[0], set_num, 0x0104, cam_ctrl->current_ae_mode) ;
            cam006_kernel_update_sceneval(&set_scene[0], set_num, 0x01E5, cam_ctrl->current_iso) ;

            cam_ctrl->set_scene_reg = &set_scene[0];
            cam_ctrl->set_scene_reg_size = set_num;

            cam_ctrl->current_wb_update = FALSE ;
            cam_ctrl->current_ae_mode_update = FALSE ;
            cam_ctrl->current_iso_update = FALSE ;

            break;
        case CAM_SCENE_PORTRAIT:

            cam_ctrl->set_scene_reg = &cam_regs.set_scene_portrait_reg[0];
            cam_ctrl->set_scene_reg_size = cam_regs.set_scene_portrait_reg_size;
            break;
        case CAM_SCENE_SCENERY:

            cam_ctrl->set_scene_reg = &cam_regs.set_scene_scenery_reg[0];
            cam_ctrl->set_scene_reg_size = cam_regs.set_scene_scenery_reg_size;
            break;
        case CAM_SCENE_NIGHT_PORTRAIT:

            cam_ctrl->set_scene_reg = &cam_regs.set_scene_night_portrait_reg[0];
            cam_ctrl->set_scene_reg_size = cam_regs.set_scene_night_portrait_reg_size;
            break;
        case CAM_SCENE_NIGHT_SCENERY:

            cam_ctrl->set_scene_reg = &cam_regs.set_scene_night_scenery_reg[0];
            cam_ctrl->set_scene_reg_size = cam_regs.set_scene_night_scenery_reg_size;
            break;
        case CAM_SCENE_ACTION:

            cam_ctrl->set_scene_reg = &cam_regs.set_scene_action_reg[0];
            cam_ctrl->set_scene_reg_size = cam_regs.set_scene_action_reg_size;
            break;
        default:
            ERR_LOG("SCENE_PARM_ERR\n");
            cam_ctrl->set_scene_reg = &cam_regs.set_scene_auto_reg[0];
            cam_ctrl->set_scene_reg_size = cam_regs.set_scene_auto_reg_size;
            break;
    }

    free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

    save_scene_mode = cam_ctrl->current_scene;
    cam_ctrl->current_scene = scene_mode;

    cam_ctrl->wait_flag = 0;

    rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_high_interrupt, IRQF_TRIGGER_RISING, "cam_vsync_h", 0);
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_set_scene cam vsync high request_irq failed\n");
        cam_ctrl->current_scene = save_scene_mode;

        cam_ctrl->current_wb_update = save_wb_update ;
        cam_ctrl->current_ae_mode_update = save_ae_mode_update ;
        cam_ctrl->current_iso_update = save_iso_update ;
        cam_ctrl->current_scene_update = TRUE ;

        return ERR_CODE;
    }

    L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_SCENE_MODE );

    rc = cam_wait_event();
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_set_scene cam_wait_event Err\n");
        cam_ctrl->current_scene = save_scene_mode;

        cam_ctrl->current_wb_update = save_wb_update ;
        cam_ctrl->current_ae_mode_update = save_ae_mode_update ;
        cam_ctrl->current_iso_update = save_iso_update ;
        cam_ctrl->current_scene_update = TRUE ;

        L_cam006_delete_vsync_high();

        return ERR_CODE;
    }

    L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

    if( cam_ctrl->scene_err_flg == TRUE )
    {

        ERR_LOG("cam006_kernel_set_scene Vsync I2C Err\n");
        cam_ctrl->current_scene = save_scene_mode;

        cam_ctrl->current_wb_update = save_wb_update ;
        cam_ctrl->current_ae_mode_update = save_ae_mode_update ;
        cam_ctrl->current_iso_update = save_iso_update ;
        cam_ctrl->current_scene_update = TRUE ;

        return ERR_CODE;
    }

    cam_ctrl->current_scene_update = FALSE ;

    switch( scene_mode )
    {
        case CAM_SCENE_AUTO:
            scene_info[ CAM_SCENE_AUTO ].wb = cam_ctrl->current_wb;
            scene_info[ CAM_SCENE_AUTO ].af_mode = cam_ctrl->af_mode;
            scene_info[ CAM_SCENE_AUTO ].ae_mode = cam_ctrl->current_ae_mode;
            scene_info[ CAM_SCENE_AUTO ].iso = cam_ctrl->current_iso;
            scene_info[ CAM_SCENE_AUTO ].flash = cam_ctrl->flash_mode;
            scene_info[ CAM_SCENE_AUTO ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_AUTO ].lens_shading = cam_ctrl->current_lens_shading;
            break;
        case CAM_SCENE_PORTRAIT:
            scene_info[ CAM_SCENE_PORTRAIT ].flash = cam_ctrl->flash_mode;
            scene_info[ CAM_SCENE_PORTRAIT ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_PORTRAIT ].lens_shading = cam_ctrl->current_lens_shading;
            break;
        case CAM_SCENE_SCENERY:
            scene_info[ CAM_SCENE_SCENERY ].flash = cam_ctrl->flash_mode;
            scene_info[ CAM_SCENE_SCENERY ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_SCENERY ].lens_shading = cam_ctrl->current_lens_shading;
            break;
        case CAM_SCENE_NIGHT_PORTRAIT:
            scene_info[ CAM_SCENE_NIGHT_PORTRAIT ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_NIGHT_PORTRAIT ].lens_shading = cam_ctrl->current_lens_shading;
            break;
        case CAM_SCENE_NIGHT_SCENERY:
            scene_info[ CAM_SCENE_NIGHT_SCENERY ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_NIGHT_SCENERY ].lens_shading = cam_ctrl->current_lens_shading;
            break;
        case CAM_SCENE_ACTION:
            scene_info[ CAM_SCENE_ACTION ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_ACTION ].lens_shading = cam_ctrl->current_lens_shading;
            break;
        default:

            scene_info[ CAM_SCENE_AUTO ].wb = cam_ctrl->current_wb;
            scene_info[ CAM_SCENE_AUTO ].af_mode = cam_ctrl->af_mode;
            scene_info[ CAM_SCENE_AUTO ].ae_mode = cam_ctrl->current_ae_mode;
            scene_info[ CAM_SCENE_AUTO ].iso = cam_ctrl->current_iso;
            scene_info[ CAM_SCENE_AUTO ].flash = cam_ctrl->flash_mode;
            scene_info[ CAM_SCENE_AUTO ].effect = cam_ctrl->current_effect;
            scene_info[ CAM_SCENE_AUTO ].lens_shading = cam_ctrl->current_lens_shading;
            break;
    }

    return rc;
}

static int cam006_kernel_get_scene_info( camera_scene_info_type* info )
{

    if( info == NULL )
    {
        ERR_LOG("cam006_kernel_get_scene_info Parm NULL Err\n");
        return ERR_CODE;
    }

    *info = scene_info[ cam_ctrl->current_scene ];

    return TRUE;

}

static int cam006_kernel_get_flash_parm_info( camera_flash_parm_type* flash_info )
{

    if( flash_info == NULL )
    {
        ERR_LOG("cam006_kernel_get_flash_parm_info Parm NULL Err\n");
        return ERR_CODE;
    }

    *flash_info = cam_ctrl->flash_parm_info;

    return TRUE;
}

static void cam006_kernel_set_flash_parm_info( camera_flash_parm_type flash_info )
{

    cam_ctrl->main_flash_cg_offset = flash_info.cg_offset;
    cam_ctrl->main_flash_sc_shift_r = flash_info.scs_r;
    cam_ctrl->main_flash_sc_shift_b = flash_info.scs_b;

    cam_ctrl->flash_parm_info.ae_diff = flash_info.ae_diff;
    cam_ctrl->flash_parm_info.awb_r_diff = flash_info.awb_r_diff;
    cam_ctrl->flash_parm_info.awb_b_diff = flash_info.awb_b_diff;
    cam_ctrl->flash_parm_info.cg_offset = flash_info.cg_offset;
    cam_ctrl->flash_parm_info.scs_r = flash_info.scs_r;
    cam_ctrl->flash_parm_info.scs_b = flash_info.scs_b;

    cam_ctrl->flash_parm_info.main_flash_on_flg = flash_info.main_flash_on_flg;

}

static int cam006_kernel_get_otp_data( camera_otp_data_type* otp_data )
{
    int rc = TRUE;

    if( otp_data == NULL )
    {
        ERR_LOG("cam006_kernel_get_otp_data Parm NULL Err\n");
        return ERR_CODE;
    }

    otp_data->o0_H = cam_ctrl->cam006_o0_H;
    otp_data->o0_M = cam_ctrl->cam006_o0_M;
    otp_data->o0_L = cam_ctrl->cam006_o0_L;
    otp_data->o1_H = cam_ctrl->cam006_o1_H;
    otp_data->o1_M = cam_ctrl->cam006_o1_M;
    otp_data->o1_L = cam_ctrl->cam006_o1_L;

    return rc;
}

static int cam006_kernel_set_af_otp_data( camera_set_af_otp_data_type af_otp_data )
{

    int rc = TRUE;

    cam_ctrl->cam006_af_otp_data.focus_optical_inf = af_otp_data.focus_optical_inf;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2541].wdata = FOCUS_MECH_INF;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2541].wdata = FOCUS_MECH_INF;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2541].wdata = FOCUS_MECH_INF;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2541].wdata  = FOCUS_MECH_INF;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2367].wdata = FOCUS_MECH_INF;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2367].wdata = FOCUS_MECH_INF;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2367].wdata = FOCUS_MECH_INF;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2367].wdata  = FOCUS_MECH_INF;

    cam_ctrl->cam006_af_otp_data.focus_10cm_macro = af_otp_data.focus_10cm_macro;

    cam_ctrl->cam006_af_otp_data.af_drv_aton_f = af_otp_data.af_drv_aton_f;
    cam_ctrl->cam006_af_otp_data.af_drv_atof_f = af_otp_data.af_drv_atof_f;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2368].wdata = cam_ctrl->cam006_af_otp_data.af_drv_aton_f;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2368].wdata = cam_ctrl->cam006_af_otp_data.af_drv_aton_f;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2368].wdata = cam_ctrl->cam006_af_otp_data.af_drv_aton_f;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2368].wdata  = cam_ctrl->cam006_af_otp_data.af_drv_aton_f;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2370].wdata = cam_ctrl->cam006_af_otp_data.af_drv_atof_f;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2370].wdata = cam_ctrl->cam006_af_otp_data.af_drv_atof_f;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2370].wdata = cam_ctrl->cam006_af_otp_data.af_drv_atof_f;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2370].wdata  = cam_ctrl->cam006_af_otp_data.af_drv_atof_f;

    cam_ctrl->cam006_af_otp_data.af_drv_aton_s = af_otp_data.af_drv_aton_s;
    cam_ctrl->cam006_af_otp_data.af_drv_atof_s = af_otp_data.af_drv_atof_s;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2369].wdata = cam_ctrl->cam006_af_otp_data.af_drv_aton_s;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2369].wdata = cam_ctrl->cam006_af_otp_data.af_drv_aton_s;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2369].wdata = cam_ctrl->cam006_af_otp_data.af_drv_aton_s;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2369].wdata  = cam_ctrl->cam006_af_otp_data.af_drv_aton_s;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2371].wdata = cam_ctrl->cam006_af_otp_data.af_drv_atof_s;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2371].wdata = cam_ctrl->cam006_af_otp_data.af_drv_atof_s;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2371].wdata = cam_ctrl->cam006_af_otp_data.af_drv_atof_s;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2371].wdata  = cam_ctrl->cam006_af_otp_data.af_drv_atof_s;

    cam_ctrl->cam006_af_otp_data.af_caf_pws = af_otp_data.af_caf_pws;
    cam_ctrl->cam006_af_otp_data.af_caf_wmm = af_otp_data.af_caf_wmm;
    cam_ctrl->cam006_af_otp_data.af_caf_cs = af_otp_data.af_caf_cs;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2406].wdata = cam_ctrl->cam006_af_otp_data.af_caf_pws;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2406].wdata = cam_ctrl->cam006_af_otp_data.af_caf_pws;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2406].wdata = cam_ctrl->cam006_af_otp_data.af_caf_pws;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2406].wdata  = cam_ctrl->cam006_af_otp_data.af_caf_pws;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2521].wdata = cam_ctrl->cam006_af_otp_data.af_caf_wmm;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2521].wdata = cam_ctrl->cam006_af_otp_data.af_caf_wmm;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2521].wdata = cam_ctrl->cam006_af_otp_data.af_caf_wmm;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2521].wdata  = cam_ctrl->cam006_af_otp_data.af_caf_wmm;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2407].wdata = cam_ctrl->cam006_af_otp_data.af_caf_cs;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2407].wdata = cam_ctrl->cam006_af_otp_data.af_caf_cs;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2407].wdata = cam_ctrl->cam006_af_otp_data.af_caf_cs;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2407].wdata  = cam_ctrl->cam006_af_otp_data.af_caf_cs;

    cam_ctrl->cam006_af_otp_data.af_alt = af_otp_data.af_alt;
    cam_ctrl->cam006_af_otp_data.af_aht = af_otp_data.af_aht;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2373].wdata = FOCUS_MECH_INF;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2373].wdata = FOCUS_MECH_INF;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2373].wdata = FOCUS_MECH_INF;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2373].wdata  = FOCUS_MECH_INF;
    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2374].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2374].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2374].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2374].wdata  = cam_ctrl->cam006_af_otp_data.af_aht;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2375].wdata = FOCUS_MECH_INF;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2375].wdata = FOCUS_MECH_INF;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2375].wdata = FOCUS_MECH_INF;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2375].wdata  = FOCUS_MECH_INF;
    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2376].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2376].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2376].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2376].wdata  = cam_ctrl->cam006_af_otp_data.af_aht;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2377].wdata = FOCUS_MECH_INF;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2377].wdata = FOCUS_MECH_INF;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2377].wdata = FOCUS_MECH_INF;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2377].wdata  = FOCUS_MECH_INF;
    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2378].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2378].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2378].wdata = cam_ctrl->cam006_af_otp_data.af_aht;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2378].wdata  = cam_ctrl->cam006_af_otp_data.af_aht;

    cam_ctrl->cam006_af_otp_data.af_sot = af_otp_data.af_sot;

    cam_regs.preview_sxga_init_reg_settings_4[cam_preview_sxga_offset.offset_2362].wdata = cam_ctrl->cam006_af_otp_data.af_sot;
    cam_regs.preview_720p_init_reg_settings_4[cam_preview_720p_offset.offset_2362].wdata = cam_ctrl->cam006_af_otp_data.af_sot;
    cam_regs.preview_wvga_init_reg_settings_4[cam_preview_wvga_offset.offset_2362].wdata = cam_ctrl->cam006_af_otp_data.af_sot;
    cam_regs.preview_vga_init_reg_settings_4[cam_preview_vga_offset.offset_2362].wdata  = cam_ctrl->cam006_af_otp_data.af_sot;

    rc = cam_i2c_write( cam_client->addr, 0x4844, cam_ctrl->cam006_af_otp_data.af_sot, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4844 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x486A, cam_ctrl->cam006_af_otp_data.focus_optical_inf, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x486A Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x486C, cam_ctrl->cam006_af_otp_data.af_drv_aton_f, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x486C Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x486E, cam_ctrl->cam006_af_otp_data.af_drv_aton_s, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x486E Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4870, cam_ctrl->cam006_af_otp_data.af_drv_atof_f, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4870 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4872, cam_ctrl->cam006_af_otp_data.af_drv_atof_s, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4872 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4876, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4876 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4878, cam_ctrl->cam006_af_otp_data.af_aht, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4878 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x487A, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x487A Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x487C, cam_ctrl->cam006_af_otp_data.af_aht, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x487C Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x487E, cam_ctrl->cam006_af_otp_data.af_alt, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x487E Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4880, cam_ctrl->cam006_af_otp_data.af_aht, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4880 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4822, cam_ctrl->cam006_af_otp_data.af_caf_pws, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4822 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4824, cam_ctrl->cam006_af_otp_data.af_caf_cs, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4824 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4838, cam_ctrl->cam006_af_otp_data.af_caf_wmm, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4838 Write Err\n");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x4852, cam_ctrl->cam006_af_otp_data.focus_optical_inf, WORD_LEN );
    if( rc < 0 )
    {
        ERR_LOG("0x4852 Write Err\n");
        return ERR_CODE;
    }

    return rc;

}

static int cam006_kernel_get_exif_info(camera_exif_info *exif)
{

  if(exif == NULL)
  {
    ERR_LOG("cam006_kernel_get_exif_info Parm NULL Err\n");
    return ERR_CODE;
  }

  exif->exposure_time    = cam_ctrl->exif_exposure_time;

  exif->f_number         = EXIF_F_NUMBER;

  if(cam_ctrl->flash_mode == 2){

    exif->flash = (EXIF_FLASH_COMP_FIRE | EXIF_FLASH_FLASH_ON);
  }
  else if(cam_ctrl->flash_mode == 1){

    exif->flash = EXIF_FLASH_AUTO_FIRE;
    if(cam_ctrl->flash_on_flg == TRUE){
      exif->flash |= EXIF_FLASH_FLASH_ON;
    }
  }
  else{

    exif->flash = EXIF_FLASH_COMP_SUPPRES;
  }

  return TRUE;

}

static int cam006_kernel_get_otp_check_result(int *result)
{

    if( result == NULL )
    {
        ERR_LOG("cam006_kernel_get_otp_check_result Parm NULL Err\n");
        return ERR_CODE;
    }

    do{
        *result = 0;

        if(cam_ctrl->cam006_o0_H != 0)
            break;
        if(cam_ctrl->cam006_o0_M != 0)
            break;
        if(cam_ctrl->cam006_o0_L != 0)
            break;
        if(cam_ctrl->cam006_o1_H != 0)
            break;
        if(cam_ctrl->cam006_o1_M != 0)
            break;
        if(cam_ctrl->cam006_o1_L != 0)
            break;

        *result = -1;

    }while(0);

    return TRUE;

}

static int cam006_kernel_set_device_mode(int mode){

    msm_batt_oem_update_dev_info_cam((dev_info_cam_state_type)mode);

    return TRUE;
}

static int cam006_kernel_prepare_snapshot( void )
{
    int rc = TRUE;

    rc = L_cam006_kernel_check_flash_on( &(cam_ctrl->flash_parm_info.flash_flg) );
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_prepare_snapshot flash_check Err\n");
        return ERR_CODE;
    }

    if( cam_ctrl->flash_parm_info.flash_flg != TRUE )
    {

        return TRUE;
    }

    cam_ctrl->still_err_flg = FALSE;

    cam_ctrl->wait_flag = 0;

    free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio),0);

    L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_NONE );

    rc = request_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), vsync_high_interrupt, IRQF_TRIGGER_RISING, "cam_vsync_h", 0);
    if (rc < 0)
    {
        ERR_LOG("cam vsync high request_irq failed!\n");
        return ERR_CODE;
    }

    L_cam006_entry_vsync_high_proc( CAMERA_V_RISE_YUV_JPEG_STILL_PRE_FLASH_ON_AF_OFF );

    rc = cam_wait_event();
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_prepare_snapshot Current Thread Wait Err\n");
        return ERR_CODE;
    }

    if( cam_ctrl->still_err_flg == TRUE )
    {
        rc = ERR_CODE;
    }
    else
    {
        rc = TRUE;
    }

    return rc;
}

static int L_cam006_kernel_still_movie_chg_retry_set_image_control(camera_power_on_reg_type type)
{
    int rc = TRUE;

    if( cam_ctrl->current_scene != CAM_SCENE_AUTO )
    {
      cam_ctrl->current_scene_update = TRUE;
    }
    else
    {
      cam_ctrl->current_scene_update = FALSE;
    }

    if( cam_ctrl->current_wb != CAM_WB_AUTO )
    {
      cam_ctrl->current_wb_update = TRUE;
    }
    else
    {
      cam_ctrl->current_wb_update = FALSE;
    }

    if(( cam_ctrl->current_iso != CAM_ISO_AUTO ) && (type == CAMERA_POWER_ON_NORMAL))
    {
      cam_ctrl->current_iso_update = TRUE;
    }
    else
    {
      cam_ctrl->current_iso_update = FALSE;
    }

    if( cam_ctrl->current_ae_mode != CAM_AEC_AVERAGE )
    {
      cam_ctrl->current_ae_mode_update = TRUE;
    }
    else
    {
      cam_ctrl->current_ae_mode_update = FALSE;
    }

    rc = cam006_kernel_set_scene(cam_ctrl->current_scene);
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_kernel_still_movie_chg_retry_set_image_control Set SCENE Err\n");
        return ERR_CODE;
    }

    rc = cam006_kernel_set_effect( cam_ctrl->current_effect );
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_kernel_still_movie_chg_retry_set_image_control Set Effect Err\n");
        return ERR_CODE;
    }

    rc = cam006_kernel_set_lens_shading( cam_ctrl->current_lens_shading );
    if( rc < 0 )
    {
        ERR_LOG("L_cam006_kernel_still_movie_chg_retry_set_image_control Set LensShading Err\n");
        return ERR_CODE;
    }

    return rc;

}

int cam_sensor_config(void __user *argp)
{
    struct sensor_cfg_data cfg_data;
    long   rc = 0;

    if (copy_from_user(
                &cfg_data,
                (void *)argp,
                sizeof(struct sensor_cfg_data)))
        return -EFAULT;

    switch (cfg_data.cfgtype) {
    case CFG_SET_MODE:

      down(&cam_sem);

      rc = cam_set_sensor_mode(cfg_data.mode, &(cfg_data.cfg.gfps));

      up(&cam_sem);
      break;

    case CFG_MOVE_FOCUS:

      down(&cam_sem);

      rc = cam_set_move_focus(cfg_data.mode);

      up(&cam_sem);
      break;

    case CFG_SET_DEFAULT_FOCUS:

      down(&cam_sem);

      rc = cam_set_default_focus();

      up(&cam_sem);
      break;

#ifdef CAMSENSOR_DEBUG
    case CFG_SET_EFFECT:

      cam_dbg_prev_mode = cfg_data.cfg.effect;

      cam_chg_state(SENSOR_STATE_MODE, SENSOR_PREV_DEBUG);

        break;
    case CFG_SET_ANTIBANDING:

      cam_dbg_flash_mode = cfg_data.cfg.effect;
      break;
    case CFG_SET_WB:

      cam_dbg_preflash_mode = cfg_data.cfg.effect;
      break;
    case CFG_SET_EXPOSURE_MODE:

      cam_dbg_videolight_mode = cfg_data.cfg.effect;
      break;
#endif

  case CFG_SET_DIMENSIONS:

    down(&cam_sem);

    rc = cam006_kernel_set_dimensions( cfg_data );
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_dimensions Err\n");
    }

    up(&cam_sem);

    break;
  case CFG_SET_STILL_TO_PREVIEW:

    down(&cam_sem);

    rc = cam006_still_to_preview();
    if( rc < 0 )
    {
      ERR_LOG("cam006_still_to_preview Err\n");
    }

    up(&cam_sem);

    break;
  case CFG_SET_AUTO_FOCUS:

    down(&cam_sem);

    cam_ctrl->af_proc_call_flg = TRUE;

    rc = L_cam006_kernel_check_flash_on( &( cam_ctrl->flash_parm_info.flash_flg ));
    if( rc < 0 )
    {
        ERR_LOG("CFG_SET_AUTO_FOCUS check_flash_on() Err\n");

        up(&cam_sem);
        break;
    }

    if( cam_ctrl->flash_parm_info.flash_flg == TRUE )

    {
        rc = cam006_kernel_flash_seq_auto_focus((camera_af_mode_type)cfg_data.mode );
        if( rc < 0 )
        {
            ERR_LOG("cam006_kernel_flash_seq_auto_focus Err\n");
        }
    }

    else
    {

        rc = cam006_start_auto_focus((uint8_t)cfg_data.mode);
        if( rc < 0 )
        {
          ERR_LOG("cam006_start_auto_focus Err\n");
        }
    }

    up(&cam_sem);

    break;
  case CFG_SET_WB:
    rc = cam006_kernel_set_wb((camera_wb_type)cfg_data.cfg.value );
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_wb Err\n");
    }
    break;
  case CFG_SET_EFFECT:
    rc = cam006_kernel_set_effect((camera_effect_type)cfg_data.cfg.value );
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_effect Err\n");
    }
    break;
  case CFG_SET_ISO:
    rc = cam006_kernel_set_iso((camera_effect_type)cfg_data.cfg.value );
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_iso Err\n");
    }
    break;
  case CFG_SET_EXPOSURE_MODE:
    rc = cam006_kernel_set_ae_mode((camera_effect_type)cfg_data.cfg.value );
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_ae_mode Err\n");
    }
    break;

  case CFG_SET_AF_WINDOW:
    rc = cam006_kernel_set_af_window(&(cfg_data.cfg.afwin_info));
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_af_window Err\n");
    }
    break;
  case CFG_SET_AE_WINDOW:
    rc = cam006_kernel_set_ae_window(&(cfg_data.cfg.aewin_info));
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_ae_window Err\n");
    }
    break;
  case CFG_SET_AWB_WINDOW:
    rc = cam006_kernel_set_awb_window(&(cfg_data.cfg.awbwin_info));
    if( rc < 0 )
    {
      ERR_LOG("cam006_kernel_set_awb_window Err\n");
    }
    break;

  case CFG_SET_LENS_SHADING:
    rc = cam006_kernel_set_lens_shading(( uint8_t )cfg_data.cfg.value);
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_set_lens_shading Err\n");
    }
    break;

  case CFG_SET_SCENE:

    down(&cam_sem);

    rc = cam006_kernel_set_scene(( camera_scene_type )cfg_data.mode );
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_set_scene Err\n");
    }

    up(&cam_sem);

    break;
  case CFG_GET_SCENE_INFO:
    rc = cam006_kernel_get_scene_info(&(cfg_data.scene_info ));
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_get_scene_info Err\n");
    }
    else
    {

        if( copy_to_user( argp, &cfg_data, sizeof( struct sensor_cfg_data )))
        {
            ERR_LOG("CFG_GET_SCENE_INFO copy_to_user() Err\n");
            rc = ERR_CODE;
        }
        else
        {
            rc = TRUE;
        }
    }
    break;

  case CFG_GET_FLASH_PARM:
    rc = cam006_kernel_get_flash_parm_info(&(cfg_data.flash_parm));
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_get_flash_parm_info Err\n");
    }
    else
    {

        if( copy_to_user( argp, &cfg_data, sizeof( struct sensor_cfg_data )))
        {
            ERR_LOG("CFG_GET_FLASH_PARM copy_to_user() Err\n");
            rc = ERR_CODE;
        }
        else
        {
            rc = TRUE;
        }
    }
    break;
  case CFG_SET_FLASH_PARM:
    cam006_kernel_set_flash_parm_info( cfg_data.flash_parm );
    break;

  case CFG_GET_EXIF_INFO:
    rc = cam006_kernel_get_exif_info(&(cfg_data.exif_info));
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_get_scene_info Err\n");
    }
    else
    {

        if( copy_to_user( argp, &cfg_data, sizeof( struct sensor_cfg_data )))
        {
            ERR_LOG("CFG_GET_EXIF_INFO copy_to_user() Err\n");
            rc = ERR_CODE;
        }
        else
        {
            rc = TRUE;
        }
    }
    break;

  case CFG_GET_OTP_CHECK_RESULT:
    rc = cam006_kernel_get_otp_check_result(&(cfg_data.otp_check_result));
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_get_otp_check_result Err\n");
    }
    else
    {

        if( copy_to_user( argp, &cfg_data, sizeof(struct sensor_cfg_data)))
        {
            ERR_LOG("cam006_kernel_get_otp_check_result copy_to_user() Err\n");
            rc = ERR_CODE;
        }
        else
        {
            rc = TRUE;
        }
    }
    break;

  case CFG_GET_OTP_DATA:
    rc = cam006_kernel_get_otp_data( &cfg_data.otp_data );
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_get_otp_data() Err\n");
    }
    else
    {

        if( copy_to_user( argp, &cfg_data, sizeof( struct sensor_cfg_data )))
        {
            ERR_LOG("CFG_GET_OTP_DATA copy_to_user() Err\n");
            rc = ERR_CODE;
        }
        else
        {
            rc = TRUE;
        }
    }
    break;

  case CFG_SET_OTP_AF_DATA:

    down(&cam_sem);

    rc = cam006_kernel_set_af_otp_data( cfg_data.af_otp_data );
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_set_af_otp_data Err\n");
    }

    up(&cam_sem);

    break;

  case CFG_PREPARE_SNAPSHOT:

    down(&cam_sem);

    rc = cam006_kernel_prepare_snapshot();
    if( rc < 0 )
    {
        ERR_LOG("CFG_PREPARE_SNAPSHOT Err\n");
    }

    up(&cam_sem);

    break;

  case CFG_SET_DEVICE_INFO:
    rc = cam006_kernel_set_device_mode((int)cfg_data.mode );
    if( rc < 0 )
    {
        ERR_LOG("cam006_kernel_set_scene Err\n");
    }
    break;

    default:
        rc = -EFAULT;
        break;
    }

    return rc;
}

int cam_sensor_release(void)
{
    int rc = 0;
  unsigned int flash_state = 0x00000000;

  down(&cam_sem);

  if(cam_ctrl->main_flash_on_flg == TRUE ||
     cam_ctrl->pre_flash_on_flg == TRUE)
  {

    cam_ctrl->main_flash_on_flg = FALSE;
    cam_ctrl->pre_flash_on_flg = FALSE;

    rc = L_cam006_kernel_flash_off();
  }

  L_cam006_kernel_sensor_power_off();

  up(&cam_sem);

  if(cam_ctrl) {

    #ifdef AF_FLASH_BB_CHECK
    gpio_direction_output( AF_CHK_GPIO, GPIO_LO );
    gpio_direction_output( FL1_IN_GPIO, GPIO_LO );
    #endif

    free_irq(gpio_to_irq(cam_ctrl->sensordata->vsync_gpio), 0);

    flash_state = cam_get_state(SENSOR_STATE_FLASH);
    if(flash_state &
        (SENSOR_FLASH_ON | SENSOR_PRE_FLASH_ON | SENSOR_VIDEO_LIGHT_ON) )
    {
      cam628_flash_disable();
    }

    tasklet_kill( &cam_tasklet_vsync_high );
    tasklet_kill( &cam_tasklet_vsync_low );
    tasklet_kill( &cam_tasklet_vsync_watch );

    cam_ctrl->camera_state = CAMERA_INIT_STATE;
    cam_ctrl->focus_state = CAMERA_IDLE_FOCUS;
    cam_ctrl->picture_size_chenge = FALSE;

      kfree(cam_ctrl);

  }

    return rc;
}

static int cam_i2c_probe(struct i2c_client *client,
    const struct i2c_device_id *id)
{
    int rc = 0;

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        rc = -ENOTSUPP;
        goto probe_failure;
    }

    cam_sensorw =
        kzalloc(sizeof(struct cam_work_t), GFP_KERNEL);

    if (!cam_sensorw) {
        rc = -ENOMEM;
        goto probe_failure;
    }

    i2c_set_clientdata(client, cam_sensorw);
    cam_init_client(client);
    cam_client = client;

    return 0;

probe_failure:
    kfree(cam_sensorw);
    cam_sensorw = NULL;

    return rc;
}

static int __exit cam_remove(struct i2c_client *client)
{
    struct cam_work_t *sensorw = i2c_get_clientdata(client);

    free_irq(client->irq, sensorw);
    cam_client = NULL;
    cam_sensorw = NULL;
    kfree(sensorw);

    return 0;
}

static const struct i2c_device_id cam_i2c_id[] = {
    { MCAM_I2C_NAME, 2},
    { },
};

static struct i2c_driver cam_i2c_driver = {
    .id_table = cam_i2c_id,
    .probe  = cam_i2c_probe,
    .remove = __exit_p(cam_remove),
    .driver = {
        .name = MCAM_I2C_NAME,
    },
};

static int cam_sensor_probe(const struct msm_camera_sensor_info *info,
                struct msm_sensor_ctrl *s)
{
    int rc = i2c_add_driver(&cam_i2c_driver);
    if (rc < 0) {
      
        rc = -ENOTSUPP;
        goto probe_done;
    }

    rc = cam_sensor_init_probe(info);
    if (rc < 0)
        goto probe_done;

    s->s_init = cam_sensor_init;
    s->s_release = cam_sensor_release;
    s->s_config  = cam_sensor_config;

probe_done:
    
    return 0;
}

static int __cam_probe(struct platform_device *pdev)
{
    return msm_camera_drv_start(pdev, cam_sensor_probe);
}

static struct platform_driver msm_camera_driver = {
    .probe = __cam_probe,
    .driver = {
        .name = MCAM_SENSOR_DEVICE_NAME,
        .owner = THIS_MODULE,
    },
};

static int __init cam_init(void)
{
    return platform_driver_register(&msm_camera_driver);
}

int cam006_initial_setting_1( camera_power_on_reg_type power_on_type )
{
    int rc = -EFAULT;

    struct cam_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    if( power_on_type == CAMERA_POWER_ON_NORMAL )
    {

        set_reg = &cam_regs.preview_sxga_init_reg_settings_1[0];
        set_reg_size = cam_regs.preview_sxga_init_reg_settings_1_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_720P )
    {

        set_reg = &cam_regs.preview_720p_init_reg_settings_1[0];
        set_reg_size = cam_regs.preview_720p_init_reg_settings_1_size;

    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_WVGA )
    {

        set_reg = &cam_regs.preview_wvga_init_reg_settings_1[0];
        set_reg_size = cam_regs.preview_wvga_init_reg_settings_1_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_VGA )
    {

        set_reg = &cam_regs.preview_vga_init_reg_settings_1[0];
        set_reg_size = cam_regs.preview_vga_init_reg_settings_1_size;
    }

    else
    {

        set_reg = &cam_regs.dlc_vga_init_reg_settings_1[0];
        set_reg_size = cam_regs.dlc_vga_init_reg_settings_1_size;
    }

    rc = cam_i2c_write_table( set_reg, set_reg_size );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    return rc;
}

int cam006_o_data_read( void )
{
    int rc = -EFAULT;

    rc = cam_i2c_read_long( cam_client->addr, 0x0250, &cam_ctrl->cam006_o0_L, ADWORD_DATLONG_LEN );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    rc = cam_i2c_read_long( cam_client->addr, 0x0254, &cam_ctrl->cam006_o0_M, ADWORD_DATLONG_LEN );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    rc = cam_i2c_read_long( cam_client->addr, 0x0258, &cam_ctrl->cam006_o0_H, ADWORD_DATLONG_LEN );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    rc = cam_i2c_read_long( cam_client->addr, 0x025C, &cam_ctrl->cam006_o1_L, ADWORD_DATLONG_LEN );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    rc = cam_i2c_read_long( cam_client->addr, 0x0260, &cam_ctrl->cam006_o1_M, ADWORD_DATLONG_LEN );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    rc = cam_i2c_read_long( cam_client->addr, 0x0264, &cam_ctrl->cam006_o1_H, ADWORD_DATLONG_LEN );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    return rc;
}

int cam006_check_internal_status( unsigned short comp_stat_data, unsigned char clr_data )
{
    int rc = TRUE;
    int i;
    unsigned short read_data = 0;

    for( i=0; i < 500; i++ )
    {

        rc = cam_i2c_read( cam_client->addr, 0x00F8, &read_data, ADWORD_DATBYTE_LEN );
        if( rc < 0 )
        {
            printk("cam_i2c_read Err:%d\n", read_data);
            return ERR_CODE;
        }

        if( comp_stat_data & read_data )
        {

            break;
        }
        mdelay( 10 );
    }

    if( i >= 500 )
    {
        printk("Sensor Status Change Err");
        return ERR_CODE;
    }

    rc = cam_i2c_write( cam_client->addr, 0x00FC, clr_data, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
        printk("cam_i2c_write Err:%d\n", rc);
    }

    return rc;

}

int cam006_initial_setting_2( camera_power_on_reg_type power_on_type )
{
    int rc = -EFAULT;

    struct cam_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    if( power_on_type == CAMERA_POWER_ON_NORMAL )
    {

        set_reg = &cam_regs.preview_sxga_init_reg_settings_2[0];
        set_reg_size = cam_regs.preview_sxga_init_reg_settings_2_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_720P )
    {

        set_reg = &cam_regs.preview_720p_init_reg_settings_2[0];
        set_reg_size = cam_regs.preview_720p_init_reg_settings_2_size;

    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_WVGA )
    {

        set_reg = &cam_regs.preview_wvga_init_reg_settings_2[0];
        set_reg_size = cam_regs.preview_wvga_init_reg_settings_2_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_VGA )
    {

        set_reg = &cam_regs.preview_vga_init_reg_settings_2[0];
        set_reg_size = cam_regs.preview_vga_init_reg_settings_2_size;
    }

    else
    {

        set_reg = &cam_regs.dlc_vga_init_reg_settings_2[0];
        set_reg_size = cam_regs.dlc_vga_init_reg_settings_2_size;
    }

    rc = cam_i2c_write_table( set_reg, set_reg_size );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    return rc;
}

int cam006_initial_setting_3( void )
{
  int rc = -EFAULT;
  unsigned short normR = 0;
  unsigned short normB = 0;
  unsigned short awbpreR = 0;
  unsigned short awbpreB = 0;
  unsigned short tmp1 = 0;
  unsigned short tmp2 = 0;

  if(( cam_ctrl->cam006_o0_M & 0x0003FFC0 ) == 0 )
  {
    return 0;
  }

  tmp1 = (unsigned short)(( cam_ctrl->cam006_o0_M & 0xFFFC0000 ) >> 18 );
  tmp2 = (unsigned short)(( cam_ctrl->cam006_o0_H & 0x00000003 ) << 14 );
  normR = ( tmp1 | tmp2 ) - 0x0600;

  rc = cam_i2c_write( cam_client->addr, 0x4A04, (unsigned long)normR, WORD_LEN );
  if( rc < 0 )
  {
    printk("0x4A04 I2C Write Err\n");
    return rc;
  }

  normB = (unsigned short)(( cam_ctrl->cam006_o0_H & 0x0003FFFC ) >> 2 );

  rc = cam_i2c_write( cam_client->addr, 0x4A06, (unsigned long)normB, WORD_LEN );
  if( rc < 0 )
  {
    printk("0x4A06 I2C Write Err\n");
    return rc;
  }

  tmp1 = (unsigned short)(( cam_ctrl->cam006_o0_L & 0xFC000000 ) >> 26 );
  tmp2 = (unsigned short)(( cam_ctrl->cam006_o0_M & 0x0000003F ) << 6 );

  awbpreR = tmp2 | tmp1;

  rc = cam_i2c_write( cam_client->addr, 0x4A08, (unsigned long)awbpreR, WORD_LEN );
  if( rc < 0 )
  {
    printk("0x4A08 I2C Write Err\n");
    return rc;
  }

  awbpreB = (unsigned short)(( cam_ctrl->cam006_o0_M & 0x0003FFC0 ) >> 6 );
  rc = cam_i2c_write( cam_client->addr, 0x4A0A, (unsigned long)awbpreB, WORD_LEN );
  if( rc < 0 )
  {
    printk("0x4A0A I2C Write Err\n");
    return rc;
  }

  return rc;
}

int cam006_initial_write_rom_correction( void )
{
#ifdef REG_C20P
    return TRUE;
#else
    int rc = -EFAULT;

    rc = cam_i2c_write_table( &cam_regs.init_write_rom_correction[0],
                                  cam_regs.init_write_rom_correction_size );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    return rc;
#endif

}

int cam006_initial_setting_4( camera_power_on_reg_type power_on_type )
{
    int rc = -EFAULT;

    struct cam_i2c_reg_conf* set_reg;
    uint16_t set_reg_size;

    if( power_on_type == CAMERA_POWER_ON_NORMAL )
    {

        set_reg = &cam_regs.preview_sxga_init_reg_settings_4[0];
        set_reg_size = cam_regs.preview_sxga_init_reg_settings_4_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_720P )
    {

        set_reg = &cam_regs.preview_720p_init_reg_settings_4[0];
        set_reg_size = cam_regs.preview_720p_init_reg_settings_4_size;

    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_WVGA )
    {

        set_reg = &cam_regs.preview_wvga_init_reg_settings_4[0];
        set_reg_size = cam_regs.preview_wvga_init_reg_settings_4_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_VGA )
    {

        set_reg = &cam_regs.preview_vga_init_reg_settings_4[0];
        set_reg_size = cam_regs.preview_vga_init_reg_settings_4_size;
    }

    else
    {

        set_reg = &cam_regs.dlc_vga_init_reg_settings_4[0];
        set_reg_size = cam_regs.dlc_vga_init_reg_settings_4_size;
    }

    rc = cam_i2c_write_table( set_reg, set_reg_size );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    return rc;
}

int cam006_AFdriver_data_write( void )
{
  int rc = -EFAULT;
  int i;
  unsigned short read_data;

  rc = cam_i2c_write_table( &cam_regs.af_driver_data[0],
                                 cam_regs.af_driver_data_size );
  if( rc < 0 )
  {
    printk("failed AF Driver cam_i2c_write_table! (rc=%d)\n", rc);
    return rc;
  }

  rc = cam_i2c_write( cam_client->addr, 0x485E, (unsigned long)0x8D27, WORD_LEN );
  if( rc < 0 )
  {
    printk("failed AF Driver CheckSum I2C Write Err\n");
    return rc;
  }

  rc = cam_i2c_write( cam_client->addr, 0x485C, (unsigned long)0x028C, WORD_LEN );
  if( rc < 0 )
  {
    printk("failed AF Driver DataSize I2C Write Err\n");
    return rc;
  }

  rc = cam_i2c_write( cam_client->addr, 0x000A, (unsigned long)0x01, WORD_LEN );
  if( rc < 0 )
  {
    printk("failed AF Driver DataSize I2C Write Err\n");
    return rc;
  }

  for( i=0; i < 1000; i++ )
  {
    rc = cam_i2c_read( cam_client->addr, 0x000A, &read_data, ADWORD_DATBYTE_LEN );
    if( rc < 0 )
    {
      printk("failed AF RAMDriver Check I2C Read Err\n");
      return rc;
    }

    if( read_data & 0x02 )
    {
      break;
    }

    mdelay(2);
  }

  if( i >= 1000 )
  {
    printk("AF RAMDriver Check Status Err\n");
    rc = -EFAULT;
  }

  return rc;

}

int cam006_initial_setting_5( camera_power_on_reg_type power_on_type )
{
    int rc = -EFAULT;

    struct cam_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    if( power_on_type == CAMERA_POWER_ON_NORMAL )
    {

        set_reg = &cam_regs.preview_sxga_init_reg_settings_5[0];
        set_reg_size = cam_regs.preview_sxga_init_reg_settings_5_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_720P )
    {

        set_reg = &cam_regs.preview_720p_init_reg_settings_5[0];
        set_reg_size = cam_regs.preview_720p_init_reg_settings_5_size;

    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_WVGA )
    {

        set_reg = &cam_regs.preview_wvga_init_reg_settings_5[0];
        set_reg_size = cam_regs.preview_wvga_init_reg_settings_5_size;
    }

    else if( power_on_type == CAMERA_POWER_ON_MOVIE_VGA )
    {

        set_reg = &cam_regs.preview_vga_init_reg_settings_5[0];
        set_reg_size = cam_regs.preview_vga_init_reg_settings_5_size;
    }

    else
    {

        set_reg = &cam_regs.dlc_vga_init_reg_settings_5[0];
        set_reg_size = cam_regs.dlc_vga_init_reg_settings_5_size;
    }

    rc = cam_i2c_write_table( set_reg, set_reg_size );
    if( rc < 0 )
    {
        printk("failed cam_i2c_write_table! (rc=%d)\n", rc);
    }

    return rc;
}

int cam006_initial_all_reg_read( void )
{
    int rc = -EFAULT;

    return rc;
}

module_init(cam_init);
